# -*- coding: utf-8 -*-
# distutils: language = c++
u"""
Created on 2017-1-1

@author: cheng.li
"""

cimport cython
from PyFin.Math.MathConstants cimport NAN
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from libc.math cimport isnan
from libc.string cimport memcpy
from libcpp.list cimport list as CList
from libcpp.string cimport string as CString
from libcpp.map cimport map as CMap
from libcpp.pair cimport pair as CPair
from cython.operator import dereference, preincrement
from PyFin.Utilities.Asserts cimport require


cdef class Deque:

    def __cinit__(self,
                  size_t window):
        self.window = window
        self.is_full = False
        self.con = <double*> PyMem_Malloc(window * sizeof(double))
        for i in range(window):
            self.con[i] = 0.
        self.start = 0
        self.count = 0

    def __dealloc__(self):
        PyMem_Free(self.con)

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef double dump(self, double value, double default=NAN):
        cdef size_t n = self.start
        cdef double* con = self.con
        cdef double popout

        if self.is_full:
            popout = con[n]
            con[n] = value
            self.start = (n + 1) % self.window
            return popout
        else:
            con[self.count] = value
            self.count += 1
            self.is_full = self.count == self.window
            return default

    cpdef list dumps(self, values):
        cdef ret_values = []
        for v in values:
            ret_values.append(self.dump(v))
        return ret_values

    cdef inline size_t size(self):
        return self.count

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef size_t idx(self, double value):
        cdef size_t i
        for i in range(self.count):
            if value == self.con[i]:
                break
        else:
            i = -1
        if i < 0:
            return -1
        else:
            i = (i - self.start + self.window) % self.window
        return i

    cdef inline bint isFull(self):
        return self.is_full

    @cython.wraparound(False)
    @cython.boundscheck(False)
    cpdef double sum(self):
        cdef double x = self.con[0]
        cdef int i
        for i in range(1, self.count):
            x += self.con[i]
        return x

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def __getitem__(self, size_t item):
        return self.con[(self.start + item) % self.window]

    def __richcmp__(Deque self, Deque other, int op):
        cdef bint flag = False
        cdef int i
        if op == 2:
           flag =  self.window == other.window \
                   and self.is_full == other.is_full \
                   and self.start == other.start

           if flag:
               for i in range(self.window):
                   if self.con[i] != other.con[i]:
                       return False
               return True
           else:
               return False

        elif op == 3:
            return not self.__richcmp__(other, 2)

    cdef void set_data(self, bytes data):
        memcpy(self.con, <char*>data, sizeof(double)*self.window)

    def __reduce__(self):
        cdef bytes data = <bytes>(<char *>self.con)[:sizeof(double)*self.window]
        return rebuild_deque, (data, self.window, self.is_full, self.start, self.count)


cpdef object rebuild_deque(bytes data, size_t window, bint is_full, size_t start, size_t count):
    c = Deque(window)
    c.set_data(data)
    c.is_full = is_full
    c.start = start
    c.count = count
    return c


cdef CString _RIGHT = "right".encode("UTF-8")
cdef CString _LEFT = "left".encode("UTF-8")
cdef CString _BOTH = "both".encode("UTF-8")
cdef CString _NEITHER = "neither".encode("UTF-8")


cdef class DiffDeque:

    def __cinit__(self,
                  window,
                  closed="right"):
        self.window = window
        self.con = CList[double]()
        self.stamps = CList[double]()
        cdef str closed_str = closed.lower()
        require(closed_str in ("left", "right", "both", "neither"),
                    ValueError,
                    "closed parameter is <{0}> which is not in the recognized formats".format(closed))
        self.closed = closed_str.encode("UTF-8")
        self.last = NAN
        self.last_stamp = NAN

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef CList[double] dump(self, double value, int stamp, double default=NAN):
        cdef CList[double] ret_values = CList[double]()
        if self.closed == _BOTH or self.closed == _LEFT:
            while self.con.size() > 0 and (stamp - self.stamps.front()) > self.window:
                ret_values.push_back(self.con.front())
                self.con.pop_front()
                self.stamps.pop_front()
        else:
            while self.con.size() > 0 and (stamp - self.stamps.front()) >= self.window:
                ret_values.push_back(self.con.front())
                self.con.pop_front()
                self.stamps.pop_front()
        if self.closed == _RIGHT or self.closed == _BOTH:
            self.con.push_back(value)
            self.stamps.push_back(stamp)
        else:
            if not isnan(self.last):
                if (self.closed == _NEITHER and (stamp - self.last_stamp) < self.window) \
                    or (self.closed == _LEFT and (stamp - self.last_stamp) <= self.window):
                    self.con.push_back(self.last)
                    self.stamps.push_back(self.last_stamp)
        self.last = value
        self.last_stamp = stamp
        return ret_values

    cpdef CList[double] dumps(self, values, stamps):
        cdef CList[double] ret_values = CList[double]()
        cdef CList[double] tmp
        for v, s in zip(values, stamps):
            tmp = self.dump(v, s)
            ret_values.merge(tmp)
        return ret_values

    cpdef size_t size(self):
        return self.con.size()

    cpdef bint isFull(self):
        if self.con.size() > 0:
            return True
        else:
            return False

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef size_t idx(self, double value):
        cdef size_t i
        cdef CList[double].iterator it
        it = self.con.begin()
        for i in range(self.con.size()):
            if value == dereference(it):
                break
            else:
                preincrement(it)
        else:
            i = -1
        return i

    @cython.wraparound(False)
    @cython.boundscheck(False)
    cpdef double sum(self):
        cdef double x = 0.0
        cdef size_t i
        cdef CList[double].iterator it
        it = self.con.begin()
        for i in range(self.con.size()):
            x += dereference(it)
            preincrement(it)
        return x

    cpdef CString close(self):
        return self.closed

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def __getitem__(self, size_t item):
        cdef size_t i
        cdef CList[double].iterator it
        it = self.con.begin()
        for i in range(self.con.size()):
            if i == item:
                return dereference(it)
            preincrement(it)

    def __richcmp__(Deque self, DiffDeque other, int op):
        cdef bint flag = False
        cdef int i
        if op == 2:
            flag = self.window == other.window \
                   and self.is_full == other.is_full \
                   and self.closed == other.closed
            if flag:
                return True
            else:
                return False

        elif op == 3:
            return not self.__richcmp__(other, 2)

    def __reduce__(self):
        return rebuild_diff_deque, (self.window, self.closed.decode("UTF-8"))

cpdef object rebuild_diff_deque(double window, str closed):
    c = DiffDeque(window, closed)
    return c


cdef class UniqueDiffDeque:

    def __cinit__(self,
                  window,
                  closed="right"):
        self.window = window
        cdef str closed_str = closed.lower()
        require(closed_str in ("left", "right", "both", "neither"),
                    ValueError,
                    "closed parameter is <{0}> which is not in the recognized formats".format(closed))
        self.closed = closed_str.encode("UTF-8")
        self.last = NAN
        self.last_stamp = NAN
        self.values_map = CMap[double, double]()
        self.stamps_map = CMap[double, double]()

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef CList[double] dump(self, double value, int stamp, double default=NAN):
        cdef CList[double] ret_values = CList[double]()
        cdef double out_value
        cdef double previous_stamp
        cdef size_t debug_int
        cdef CMap[double, double].iterator it
        cdef CPair[double, double] pair_value

        it = self.stamps_map.begin()
        if self.closed == _BOTH or self.closed == _LEFT:
            while it != self.stamps_map.end() and (stamp - dereference(it).first) > self.window:
                out_value =  dereference(it).second
                ret_values.push_back(out_value)
                self.stamps_map.erase(it)
                debug_int = self.values_map.erase(out_value)
                it = self.stamps_map.begin()
        else:
            while it != self.stamps_map.end() and (stamp - dereference(it).first) >= self.window:
                out_value = dereference(it).second
                ret_values.push_back(out_value)
                self.stamps_map.erase(it)
                debug_int = self.values_map.erase(out_value)
                it = self.stamps_map.begin()

        if self.closed == _RIGHT or self.closed == _BOTH:
            it = self.values_map.find(value)
            if it != self.values_map.end():
                previous_stamp = dereference(it).second
                self.stamps_map.erase(previous_stamp)
                self.values_map[value] = stamp
            else:
                pair_value = (value, stamp)
                self.values_map.insert(pair_value)
            pair_value = (stamp, value)
            self.stamps_map.insert(pair_value)
        else:
            if not isnan(self.last):
                if (self.closed == _NEITHER and (stamp - self.last_stamp) < self.window) \
                        or (self.closed == _LEFT and (stamp - self.last_stamp) <= self.window):
                    it = self.values_map.find(self.last)
                    if it != self.values_map.end():
                        previous_stamp = dereference(it).second
                        self.stamps_map.erase(previous_stamp)
                        self.values_map[self.last] = self.last_stamp
                    else:
                        pair_value = (self.last, self.last_stamp)
                        self.values_map.insert(pair_value)
                    pair_value = (self.last_stamp, self.last)
                    self.stamps_map.insert(pair_value)
        self.last = value
        self.last_stamp = stamp
        return ret_values

    cpdef CList[double] dumps(self, values, stamps):
        cdef CList[double] ret_values = CList[double]()
        cdef CList[double] tmp
        for v, s in zip(values, stamps):
            tmp = self.dump(v, s)
            ret_values.merge(tmp)
        return ret_values

    cpdef size_t size(self):
        return self.values_map.size()

    cpdef bint isFull(self):
        if not self.values_map.empty():
            return True
        else:
            return False

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef size_t idx(self, double value):
        cdef size_t i
        cdef double k
        cdef CMap[double, double].iterator it

        i = 0
        it = self.stamps_map.begin()
        while it != self.stamps_map.end():
            if dereference(it).second == value:
                return i
            i += 1
            preincrement(it)

    @cython.wraparound(False)
    @cython.boundscheck(False)
    cpdef double sum(self):
        cdef double x = 0.0
        cdef double k
        cdef CMap[double, double].iterator it

        it = self.stamps_map.begin()
        while it != self.stamps_map.end():
            x += dereference(it).second
            preincrement(it)
        return x

    cpdef CString close(self):
        return self.closed

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def __getitem__(self, size_t item):
        cdef size_t i
        cdef double k
        cdef CMap[double, double].iterator it

        i = 0
        it = self.stamps_map.begin()
        while it != self.stamps_map.end():
            if i == item:
                return dereference(it).second
            i += 1
            preincrement(it)

    def __richcmp__(Deque self, UniqueDiffDeque other, int op):
        cdef bint flag = False
        cdef int i
        if op == 2:
            flag = self.window == other.window \
                   and self.is_full == other.is_full \
                   and self.closed == other.closed
            if flag:
                return True
            else:
                return False

        elif op == 3:
            return not self.__richcmp__(other, 2)

    def __reduce__(self):
        return rebuild_unique_diff_deque, (self.window, self.closed.decode("UTF-8"))


cpdef object rebuild_unique_diff_deque(double window, str closed):
    c = UniqueDiffDeque(window, closed)
    return c


# -*- coding: utf-8 -*-
u"""
Created on 2021-1-12

@author: cheng.li
"""

import datetime as dt
import pandas as pd
import numpy as np
from PyFin.Math.Accumulators import (
    TimeMovingCount,
    TimeMovingCountUnique
)
from PyFin.Analysis.TechnicalAnalysis import (
    SecurityTimeMovingCount,
    SecurityTimeMovingCountUnique
)

m = 14

index = pd.date_range(dt.datetime(2021, 1, 1), dt.datetime(2021, 1, 1) + dt.timedelta(days=m-1), freq="s")

df = pd.DataFrame(np.random.randint(0, 10, (len(index), 3)), columns=['x', 'y', 'z'], index=index)
df["c"] = np.random.randint(0, int(len(index) / 14), len(index))
df["stamp"] = pd.DatetimeIndex(df.index).astype(np.int64) / 1e9
print(f"df.shape: {df.shape}")

start = dt.datetime.now()
exp = SecurityTimeMovingCountUnique("7D", "x")
res1 = exp.transform(df, "factor", category_field="c")
print("Finance-Python (rolling count): {0}s".format(dt.datetime.now() - start))

start = dt.datetime.now()
grouped = df.groupby('c')
for n, g in grouped:
    exp = TimeMovingCountUnique("7D", "x")
    exp.transform(g)
print("Finance-Python using groupby (rolling count): {0}s".format(dt.datetime.now() - start))

start = dt.datetime.now()
res2 = df.groupby('c')['x'].rolling("7D").apply(lambda x: len(np.unique(x)), raw=True)
print("Pandas (rolling count): {0}s".format(dt.datetime.now() - start))

res2 = res2.sort_index(level=[1, 0]).reset_index()
np.testing.assert_array_almost_equal(res1["factor"].values, res2["x"].values)


# -*- coding: utf-8 -*-
u"""
Created on 2015-7-13

@author: cheng.li
"""

import unittest
import copy
import tempfile
import os
import pickle
from PyFin.DateUtilities import Date
from PyFin.DateUtilities import Calendar
from PyFin.DateUtilities import Period
from PyFin.Enums import BizDayConventions
from PyFin.Enums import Months
from PyFin.Enums import Weekdays


class TestCalendar(unittest.TestCase):
    def testWrongInputOfHolidayCenter(self):
        with self.assertRaises(ValueError):
            _ = Calendar('NulCalendar')

    def testCalendarConstructionIsInsensitiveOfCase(self):
        cal1 = Calendar('NullCalendar')
        cal2 = Calendar('nullcalendar')
        shouldBeTrue = cal1 == cal2
        self.assertTrue(shouldBeTrue)

        caltoBeDifferent = Calendar('China.SSE')
        shouldBeFalse = cal1 == caltoBeDifferent
        self.assertFalse(shouldBeFalse)

    def testBasicFunctions(self):

        testDate = Date(2015, 7, 11)
        cal = Calendar('China.SSE')
        self.assertTrue(cal.isWeekEnd(testDate.weekday()), "{0} is expected to be a weekend".format(testDate))
        testDate = Date(2015, 7, 13)
        self.assertTrue(not cal.isWeekEnd(testDate.weekday()), "{0} is expected not to be a weekend".format(testDate))

        testDate = Date(2015, 5, 29)
        cal = Calendar('China.SSE')
        self.assertTrue(cal.isEndOfMonth(testDate), "{0} is expected to be a end of month".format(testDate))

        testDate = Date(2015, 5, 1)
        cal = Calendar('China.SSE')
        endOfMonth = cal.endOfMonth(testDate)
        self.assertEqual(endOfMonth, Date(2015, 5, 29),
                         "The month end of 2015/5 is expected to be {0}".format(Date(2015, 5, 29)))

        bizDates1 = cal.bizDaysBetween(Date(2015, 1, 1), Date(2015, 12, 31), True, False)
        bizDates2 = cal.bizDaysBetween(Date(2015, 12, 31), Date(2015, 1, 1), False, True)
        self.assertEqual(bizDates1, bizDates2)

    def testNullCalendar(self):
        cal = Calendar("Null")

        testDate = Date(2015, 1, 1)
        self.assertTrue(cal.isBizDay(testDate))
        self.assertTrue(not cal.isHoliday(testDate))
        self.assertTrue(cal.isWeekEnd(Weekdays.Saturday))
        self.assertTrue(cal.isWeekEnd(Weekdays.Sunday))
        self.assertTrue(not cal.isWeekEnd(Weekdays.Friday))

    def testChinaSSE(self):
        # China Shanghai Securities Exchange holiday list in the year 2014
        expectedHol = [Date(2014, 1, 1), Date(2014, 1, 31),
                       Date(2014, 2, 3), Date(2014, 2, 4), Date(2014, 2, 5), Date(2014, 2, 6),
                       Date(2014, 4, 7),
                       Date(2014, 5, 1), Date(2014, 5, 2),
                       Date(2014, 6, 2),
                       Date(2014, 9, 8),
                       Date(2014, 10, 1), Date(2014, 10, 2), Date(2014, 10, 3), Date(2014, 10, 6), Date(2014, 10, 7),
                       # China Shanghai Securities Exchange holiday list in the year 2015
                       Date(2015, 1, 1), Date(2015, 1, 2),
                       Date(2015, 2, 18), Date(2015, 2, 19), Date(2015, 2, 20), Date(2015, 2, 23), Date(2015, 2, 24),
                       Date(2015, 4, 6),
                       Date(2015, 5, 1),
                       Date(2015, 6, 22),
                       Date(2015, 9, 3), Date(2015, 9, 4),
                       Date(2015, 10, 1), Date(2015, 10, 2), Date(2015, 10, 5), Date(2015, 10, 6), Date(2015, 10, 7),
                       # China Shanghai Securities Exchange holiday list in the year 2016
                       Date(2016, 1, 1),
                       Date(2016, 2, 8), Date(2016, 2, 9), Date(2016, 2, 10), Date(2016, 2, 11), Date(2016, 2, 12),
                       Date(2016, 4, 4),
                       Date(2016, 5, 2),
                       Date(2016, 6, 9), Date(2016, 6, 10),
                       Date(2016, 9, 15), Date(2016, 9, 16),
                       Date(2016, 10, 3), Date(2016, 10, 4), Date(2016, 10, 5), Date(2016, 10, 6), Date(2016, 10, 7),
                       # China Shanghai Securities Exchange holiday list in the year 2017
                       Date(2017, 1, 1), Date(2017, 1, 2),
                       Date(2017, 1, 27), Date(2017, 1, 28), Date(2017, 1, 29), Date(2017, 1, 30), Date(2017, 1, 31),
                       Date(2017, 2, 1), Date(2017, 2, 2),
                       Date(2017, 4, 2), Date(2017, 4, 3), Date(2017, 4, 4),
                       Date(2017, 5, 1),
                       Date(2017, 5, 28), Date(2017, 5, 29), Date(2017, 5, 30),
                       Date(2017, 10, 1), Date(2017, 10, 2), Date(2017, 10, 3), Date(2017, 10, 4), Date(2017, 10, 5),
                       Date(2017, 10, 6), Date(2017, 10, 7), Date(2017, 10, 8),
                       # China Shanghai Securities Exchange holiday list in the year 2018
                       Date(2018, 1, 1),
                       Date(2018, 2, 15), Date(2018, 2, 16), Date(2018, 2, 17), Date(2018, 2, 18), Date(2018, 2, 19),
                       Date(2018, 2, 20), Date(2018, 2, 21),
                       Date(2018, 4, 5), Date(2018, 4, 6), Date(2018, 4, 7),
                       Date(2018, 4, 29), Date(2018, 4, 30), Date(2018, 5, 1),
                       Date(2018, 6, 16), Date(2018, 6, 17), Date(2018, 6, 18),
                       Date(2018, 9, 22), Date(2018, 9, 23), Date(2018, 9, 24),
                       Date(2018, 10, 1), Date(2018, 10, 2), Date(2018, 10, 3), Date(2018, 10, 4), Date(2018, 10, 5),
                       Date(2018, 10, 6), Date(2018, 10, 7),
                       # China Shanghai Securities Exchange holiday list in the year 2019
                       Date(2019, 1, 1),
                       Date(2019, 2, 4), Date(2019, 2, 5), Date(2019, 2, 6), Date(2019, 2, 7), Date(2019, 2, 8),
                       Date(2019, 4, 5),
                       Date(2019, 5, 1), Date(2019, 5, 2), Date(2019, 5, 3),
                       Date(2019, 6, 7),
                       Date(2019, 9, 13),
                       Date(2019, 10, 1), Date(2019, 10, 2), Date(2019, 10, 3), Date(2019, 10, 4), Date(2019, 10, 5),
                       Date(2019, 10, 6), Date(2019, 10, 7),
                       # China Shanghai Securities Exchange holiday list in the year 2020
                       Date(2020, 1, 1),
                       Date(2020, 1, 24), Date(2020, 1, 25), Date(2020, 1, 26), Date(2020, 1, 27), Date(2020, 1, 28), Date(2020, 1, 29), Date(2020, 1, 30), Date(2020, 1, 31),
                       Date(2020, 4, 4), Date(2020, 4, 5), Date(2020, 4, 6),
                       Date(2020, 5, 1), Date(2020, 5, 2), Date(2020, 5, 3), Date(2020, 5, 4), Date(2020, 5, 5),
                       Date(2020, 6, 25), Date(2020, 6, 26), Date(2020, 6, 27),
                       Date(2020, 10, 1), Date(2020, 10, 2), Date(2020, 10, 3), Date(2020, 10, 4), Date(2020, 10, 5), Date(2020, 10, 6), Date(2020, 10, 7), Date(2020, 10, 8),
                       # China Shanghai Securities Exchange holiday list in the year 2021
                       Date(2021, 1, 1), Date(2021, 1, 2), Date(2021, 1, 3),
                       Date(2021, 2, 11), Date(2021, 2, 12), Date(2021, 2, 13), Date(2021, 2, 14), Date(2021, 2, 15), Date(2021, 2, 16), Date(2021, 2, 17),
                       Date(2021, 4, 3), Date(2021, 4, 4), Date(2021, 4, 5),
                       Date(2021, 5, 1), Date(2021, 5, 2), Date(2021, 5, 3), Date(2021, 5, 4), Date(2021, 5, 5),
                       Date(2021, 6, 12), Date(2021, 6, 13), Date(2021, 6, 14),
                       Date(2021, 9, 19), Date(2021, 9, 20), Date(2021, 9, 21),
                       Date(2021, 10, 1), Date(2021, 10, 2), Date(2021, 10, 3), Date(2021, 10, 4), Date(2021, 10, 5), Date(2021, 10, 6), Date(2021, 10, 7)
                       ]

        cal = Calendar('China.SSE')

        for day in expectedHol:
            self.assertEqual(cal.isHoliday(day), True, "{0} is expected to be a holiday in {1}".format(day, cal))
            self.assertEqual(cal.isBizDay(day), False,
                             "{0} is not expected to be a working day in {1} ".format(day, cal))

    def testChinaIB(self):

        # China Inter Bank working weekend list in the year 2014
        expectedWorkingWeekEnd = [Date(2014, 1, 26),
                                  Date(2014, 2, 8),
                                  Date(2014, 5, 4),
                                  Date(2014, 9, 28),
                                  Date(2014, 10, 11),
                                  # China Inter Bank working weekend list in the year 2015
                                  Date(2015, 1, 4),
                                  Date(2015, 2, 15),
                                  Date(2015, 2, 28),
                                  Date(2015, 9, 6),
                                  Date(2015, 10, 10),
                                  # China Inter Bank working weekend list in the year 2016
                                  Date(2016, 2, 6),
                                  Date(2016, 2, 14),
                                  Date(2016, 6, 12),
                                  Date(2016, 9, 18),
                                  Date(2016, 10, 8),
                                  Date(2016, 10, 9),
                                  # China Inter Bank working weekend list in the year 2017
                                  Date(2017, 1, 22),
                                  Date(2017, 2, 4),
                                  Date(2017, 4, 1),
                                  Date(2017, 5, 27),
                                  Date(2017, 9, 30),
                                  # China Inter Bank working weekend list in the year 2018
                                  Date(2018, 2, 11),
                                  Date(2018, 2, 24),
                                  Date(2018, 4, 8),
                                  Date(2018, 4, 28),
                                  Date(2018, 9, 29),
                                  Date(2018, 9, 30),
                                  # China Inter Bank working weekend list in the year 2019
                                  Date(2019, 2, 2),
                                  Date(2019, 2, 3),
                                  Date(2019, 4, 28),
                                  Date(2019, 5, 5),
                                  Date(2019, 9, 29),
                                  Date(2019, 10, 12),
                                  # China Inter Bank working weekend list in the year 2020
                                  Date(2020, 1, 19),
                                  Date(2020, 4, 26),
                                  Date(2020, 5, 9),
                                  Date(2020, 6, 28),
                                  Date(2020, 9, 27),
                                  Date(2020, 10, 10),
                                  # China Inter Bank working weekend list in the year 2021
                                  Date(2021, 2, 7),
                                  Date(2021, 2, 20),
                                  Date(2021, 4, 25),
                                  Date(2021, 5, 8),
                                  Date(2021, 9, 18),
                                  Date(2021, 9, 26),
                                  Date(2021, 10, 9)]

        cal = Calendar('China.IB')

        for day in expectedWorkingWeekEnd:
            self.assertEqual(cal.isHoliday(day), False, "{0} is not expected to be a holiday in {1}".format(day, cal))
            self.assertEqual(cal.isBizDay(day), True, "{0} is expected to be a working day in {1} ".format(day, cal))

    def testAdjustDate(self):
        # April 30, 2005 is a working day under IB, but a holiday under SSE
        referenceDate = Date(2005, Months.April, 30)

        sseCal = Calendar('China.SSE')
        ibCal = Calendar('China.IB')

        bizDayConv = BizDayConventions.Unadjusted
        self.assertEqual(sseCal.adjustDate(referenceDate, bizDayConv), referenceDate)
        self.assertEqual(ibCal.adjustDate(referenceDate, bizDayConv), referenceDate)

        bizDayConv = BizDayConventions.Following
        self.assertEqual(sseCal.adjustDate(referenceDate, bizDayConv), Date(2005, Months.May, 9))
        self.assertEqual(ibCal.adjustDate(referenceDate, bizDayConv), Date(2005, Months.April, 30))

        bizDayConv = BizDayConventions.ModifiedFollowing
        self.assertEqual(sseCal.adjustDate(referenceDate, bizDayConv), Date(2005, Months.April, 29))
        self.assertEqual(ibCal.adjustDate(referenceDate, bizDayConv), Date(2005, Months.April, 30))

    def testAdvanceDate(self):
        referenceDate = Date(2014, 1, 31)
        sseCal = Calendar('China.SSE')
        ibCal = Calendar('China.IB')

        bizDayConv = BizDayConventions.Following

        # test null period
        self.assertEqual(sseCal.advanceDate(referenceDate, Period('0b'), bizDayConv), Date(2014, 2, 7))

        # test negative period
        self.assertEqual(sseCal.advanceDate(referenceDate, Period('-5b'), bizDayConv), Date(2014, 1, 24))

        # The difference is caused by Feb 8 is SSE holiday but a working day for IB market
        self.assertEqual(sseCal.advanceDate(referenceDate, Period('2b'), bizDayConv), Date(2014, 2, 10))
        self.assertEqual(sseCal.advanceDate(referenceDate, Period('2d'), bizDayConv), Date(2014, 2, 7))
        self.assertEqual(ibCal.advanceDate(referenceDate, Period('2b'), bizDayConv), Date(2014, 2, 8))
        self.assertEqual(ibCal.advanceDate(referenceDate, Period('2d'), bizDayConv), Date(2014, 2, 7))

        bizDayConv = BizDayConventions.ModifiedFollowing
        # May 31, 2014 is a holiday
        self.assertEqual(sseCal.advanceDate(referenceDate, Period('4m'), bizDayConv, True), Date(2014, 5, 30))

    def testDatesList(self):

        fromDate = Date(2014, 1, 31)
        toDate = Date(2014, 2, 28)
        sseCal = Calendar('China.SSE')
        ibCal = Calendar('China.IB')

        benchmarkHol = [Date(2014, 1, 31), Date(2014, 2, 3), Date(2014, 2, 4), Date(2014, 2, 5), Date(2014, 2, 6)]
        sseHolList = sseCal.holDatesList(fromDate, toDate, False)
        self.assertEqual(sseHolList, benchmarkHol)
        ibHolList = ibCal.holDatesList(fromDate, toDate, False)
        self.assertEqual(ibHolList, benchmarkHol)

        sseHolList = sseCal.holDatesList(fromDate, toDate, True)
        benchmarkHol = [Date(2014, 1, 31), Date(2014, 2, 1), Date(2014, 2, 2), Date(2014, 2, 3), Date(2014, 2, 4),
                        Date(2014, 2, 5), Date(2014, 2, 6), Date(2014, 2, 8), Date(2014, 2, 9), Date(2014, 2, 15),
                        Date(2014, 2, 16), Date(2014, 2, 22), Date(2014, 2, 23)]
        self.assertEqual(sseHolList, benchmarkHol)
        ibHolList = ibCal.holDatesList(fromDate, toDate, True)
        benchmarkHol = [Date(2014, 1, 31), Date(2014, 2, 1), Date(2014, 2, 2), Date(2014, 2, 3), Date(2014, 2, 4),
                        Date(2014, 2, 5), Date(2014, 2, 6), Date(2014, 2, 9), Date(2014, 2, 15), Date(2014, 2, 16),
                        Date(2014, 2, 22), Date(2014, 2, 23)]
        self.assertEqual(ibHolList, benchmarkHol)

        sseWorkingDayList = sseCal.bizDatesList(fromDate, toDate)
        d = fromDate
        while d <= toDate:
            if sseCal.isBizDay(d):
                self.assertTrue(d in sseWorkingDayList and d not in sseHolList)
            d += 1

        ibWorkingDayList = ibCal.bizDatesList(fromDate, toDate)
        d = fromDate
        while d <= toDate:
            if ibCal.isBizDay(d):
                self.assertTrue(d in ibWorkingDayList and d not in ibHolList)
            d += 1

    def testCalendarWithDayConvention(self):
        sseCal = Calendar('China.SSE')

        referenceDate = Date(2015, 2, 14)
        testDate = sseCal.adjustDate(referenceDate, BizDayConventions.HalfMonthModifiedFollowing)
        self.assertEqual(testDate, Date(2015, 2, 13))

        referenceDate = Date(2014, 2, 4)
        testDate = sseCal.adjustDate(referenceDate, BizDayConventions.ModifiedPreceding)
        self.assertEqual(testDate, Date(2014, 2, 7))

        referenceDate = Date(2014, 2, 3)
        testDate = sseCal.adjustDate(referenceDate, BizDayConventions.Nearest)
        self.assertEqual(testDate, Date(2014, 2, 7))

        referenceDate = Date(2014, 2, 2)
        testDate = sseCal.adjustDate(referenceDate, BizDayConventions.Nearest)
        self.assertEqual(testDate, Date(2014, 1, 30))

        with self.assertRaises(ValueError):
            _ = sseCal.adjustDate(referenceDate, -1)

    def testCalendarDeepCopy(self):
        sseCal = Calendar('China.SSE')
        copiedCal = copy.deepcopy(sseCal)

        self.assertEqual(sseCal, copiedCal)

    def testCalendarPickle(self):
        sseCal = Calendar('China.SSE')

        f = tempfile.NamedTemporaryFile('w+b', delete=False)
        pickle.dump(sseCal, f)
        f.close()

        with open(f.name, 'rb') as f2:
            pickledCal = pickle.load(f2)
            self.assertEqual(sseCal, pickledCal)

        os.unlink(f.name)



# -*- coding: utf-8 -*-
u"""
Created on 2017-4-3

@author: cheng.li
"""


cdef double DBL_MIN
cdef double QL_EPSILON
cdef double M_SQRT2
cdef double M_SQRT_2
cdef double M_SQRTPI
cdef double M_1_SQRTPI
cdef double M_PI
cdef double NAN


# -*- coding: utf-8 -*-
u"""
Created on 2017-1-3

@author: cheng.li
"""

import warnings
from libc.math cimport fabs
from libc.math cimport fmax


cpdef int require(condition, exception, str msg="") except -1:
    if not condition:
        raise exception(msg)
    return 0


cpdef int ensureRaise(exception, str msg="") except -1:
    raise exception(msg)


cpdef int warning(condition, warn_type, str msg=""):
    if not condition:
        warnings.warn(msg, warn_type)
    return 0

cpdef bint isClose(double a, double b=0., double rel_tol=1e-09, double abs_tol=1e-12) nogil:
    return fabs(a-b) <= fmax(rel_tol * fmax(fabs(a), fabs(b)), abs_tol)



# -*- coding: utf-8 -*-
u"""
Created on 2015-7-13

@author: cheng.li
"""

import unittest
import datetime as dt
from PyFin.DateUtilities.Date import Date
from PyFin.DateUtilities.Calendar import Calendar
from PyFin.DateUtilities.Period import Period
from PyFin.Enums.BizDayConventions import BizDayConventions
from PyFin.api.DateUtilities import datesList
from PyFin.api.DateUtilities import isBizDay
from PyFin.api.DateUtilities import bizDatesList
from PyFin.api.DateUtilities import holDatesList
from PyFin.api.DateUtilities import advanceDate
from PyFin.api.DateUtilities import advanceDateByCalendar
from PyFin.api.DateUtilities import adjustDateByCalendar


class TestDateUtilities(unittest.TestCase):
    def setUp(self):
        self.fromDate = dt.datetime(2010, 1, 1)
        self.toDate = dt.datetime.today()

    def testIsBizDay(self):
        bizDates = bizDatesList('China.SSE', self.fromDate, self.toDate)

        for i, date in enumerate(bizDates):
            isBizFlag = isBizDay('China.SSE', date)
            self.assertTrue(isBizFlag, "at index {0:d} {1} is expected to be business day.".format(i, date))

        holDates = holDatesList('China.SSE', self.fromDate, self.toDate)
        for i, date in enumerate(holDates):
            isBizFlag = isBizDay('China.SSE', date)
            self.assertTrue(isBizFlag is False, "at index {0:d} {1} is expected to be business day.".format(i, date))

    def testDatesList(self):
        dtList = datesList(self.fromDate, self.toDate)
        fromDate = Date.fromDateTime(self.fromDate)

        for i, date in enumerate(dtList):
            expected = date
            calculated = (fromDate + i).toDateTime()
            self.assertEqual(date, (fromDate + i).toDateTime(), "at index {0:d}"
                                                                "expected:   {1}"
                                                                "calculated: {2}".format(i, expected, calculated))

    def testBizDatesList(self):
        bizDtList = bizDatesList('China.SSE', self.fromDate, self.toDate)
        fromDate = Date.fromDateTime(self.fromDate)
        currentDate = fromDate.toDateTime()
        cal = Calendar('China.SSE')
        while currentDate <= self.toDate:
            finpyDate = Date.fromDateTime(currentDate)
            if cal.isBizDay(finpyDate):
                self.assertTrue(finpyDate.toDateTime() in bizDtList,
                                "{0} is expected as a business day in {1}".format(finpyDate, cal))
            currentDate = (finpyDate + 1).toDateTime()

    def testHolDatesList(self):
        holDtList = holDatesList('China.SSE', self.fromDate, self.toDate)
        fromDate = Date.fromDateTime(self.fromDate)
        currentDate = fromDate.toDateTime()
        cal = Calendar('China.SSE')
        while currentDate <= self.toDate:
            finpyDate = Date.fromDateTime(currentDate)
            if not cal.isBizDay(finpyDate):
                self.assertTrue(finpyDate.toDateTime() in holDtList,
                                "{0} is expected as a holiday in {1}".format(finpyDate, cal))
            currentDate = (finpyDate + 1).toDateTime()

    def testAdvanceDate(self):
        referenceDate = dt.date.today()

        expected = advanceDate(referenceDate, '3M')
        calculated = Date.fromDateTime(referenceDate) + '3M'
        self.assertEqual(expected, calculated.toDateTime())

        expected = advanceDate(referenceDate, '-3M')
        calculated = Date.fromDateTime(referenceDate) - '3M'
        self.assertEqual(expected, calculated.toDateTime())

    def testAdjustDateByCalendar(self):
        referenceDate = Date(2014, 10, 1)
        cal = Calendar('China.SSE')

        expected = adjustDateByCalendar('China.SSE', referenceDate.toDateTime())
        calculated = cal.adjustDate(referenceDate)
        self.assertEqual(expected, calculated.toDateTime())

        referenceDate = Date(2014, 10, 8)
        expected = adjustDateByCalendar('China.SSE', referenceDate.toDateTime(), BizDayConventions.Preceding)
        calculated = cal.adjustDate(referenceDate, BizDayConventions.Preceding)
        self.assertEqual(expected, calculated.toDateTime())

    def testAdvanceDateByCalendar(self):
        referenceDate = Date(2014, 10, 1)
        cal = Calendar('China.SSE')

        expected = advanceDateByCalendar('China.SSE', referenceDate.toDateTime(), '2D')
        calculated = cal.advanceDate(referenceDate, Period('2D'))
        self.assertEqual(expected, calculated.toDateTime())

        expected = advanceDateByCalendar('China.SSE', referenceDate.toDateTime(), '2D')
        calculated = cal.advanceDate(referenceDate, Period('2D'))
        self.assertEqual(expected, calculated.toDateTime())

        expected = advanceDateByCalendar('China.SSE', referenceDate.toDateTime(), '2B')
        calculated = cal.advanceDate(referenceDate, Period('2B'))
        self.assertEqual(expected, calculated.toDateTime())

        expected = advanceDateByCalendar('China.SSE', referenceDate.toDateTime(), '1Y')
        calculated = cal.advanceDate(referenceDate, Period('1Y'))
        self.assertEqual(expected, calculated.toDateTime())



# -*- coding: utf-8 -*-
u"""
Created on 2016-1-13

@author: cheng.li
"""

import unittest
import math
import numpy as np
import copy
import pickle
import tempfile
import os
from scipy.stats import norm
from PyFin.Analysis.SecurityValueHolders import SecurityLatestValueHolder
from PyFin.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityDiffValueHolder
from PyFin.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySignValueHolder
from PyFin.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAverageValueHolder
from PyFin.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityXAverageValueHolder
from PyFin.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityExpValueHolder
from PyFin.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityLogValueHolder
from PyFin.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityPowValueHolder
from PyFin.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySqrtValueHolder
from PyFin.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAbsValueHolder
from PyFin.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityMACDValueHolder
from PyFin.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityMaximumValueHolder
from PyFin.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityMinimumValueHolder
from PyFin.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityNormInvValueHolder
from PyFin.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityCeilValueHolder
from PyFin.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityFloorValueHolder
from PyFin.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityRoundValueHolder


class TestStatelessTechnicalAnalysis(unittest.TestCase):
    def setUp(self):
        # preparing market data
        np.random.seed(0)
        aaplClose = np.random.randn(1000)
        aaplOpen = np.random.randn(1000)
        self.aapl = {'close': aaplClose, 'open': aaplOpen}

        ibmClose = np.random.randn(1000)
        ibmOpen = np.random.randn(1000)
        self.ibm = {'close': ibmClose, 'open': ibmOpen}
        self.dataSet = {'aapl': self.aapl, 'ibm': self.ibm}

    def testSecuritySignValueHolder(self):
        sign = SecuritySignValueHolder('close')

        expected = {}
        for i in range(len(self.aapl['close'])):
            data = dict(aapl=dict(close=self.aapl['close'][i],
                                  open=self.aapl['open'][i]),
                        ibm=dict(close=self.ibm['close'][i],
                                 open=self.ibm['open'][i]))
            sign.push(data)

            value = sign.value
            for name in value.index():
                expected[name] = np.sign(self.dataSet[name]['close'][i])
                calculated = value[name]
                self.assertAlmostEqual(expected[name], calculated, 12, 'at index {0}\n'
                                                                       'expected:   {1:.12f}\n'
                                                                       'calculated: {2:.12f}'
                                       .format(i, expected[name], calculated))

    def testSecurityAverageValueHolder(self):
        av = SecurityAverageValueHolder('close')
        expected = {}
        for i in range(len(self.aapl['close'])):
            data = dict(aapl=dict(close=self.aapl['close'][i],
                                  open=self.aapl['open'][i]),
                        ibm=dict(close=self.ibm['close'][i],
                                 open=self.ibm['open'][i]))
            av.push(data)

            value = av.value
            for name in value.index():
                expected[name] = np.mean(self.dataSet[name]['close'][:i+1])
                calculated = value[name]
                self.assertAlmostEqual(expected[name], calculated, 12, 'at index {0}\n'
                                                                       'expected:   {1:.12f}\n'
                                                                       'calculated: {2:.12f}'
                                       .format(i, expected[name], calculated))

    def testSecurityXAverageValueHolder(self):
        window = 10
        xav = SecurityXAverageValueHolder(window, 'close')
        exp_weight = 2.0 / (1.0 + window)

        expected = {}
        for i in range(len(self.aapl['close'])):
            data = dict(aapl=dict(close=self.aapl['close'][i],
                                  open=self.aapl['open'][i]),
                        ibm=dict(close=self.ibm['close'][i],
                                 open=self.ibm['open'][i]))
            xav.push(data)

            value = xav.value
            for name in value.index():
                if i == 0:
                    expected[name] = self.dataSet[name]['close'][i]
                else:
                    expected[name] += exp_weight * (self.dataSet[name]['close'][i] - expected[name])
                calculated = value[name]
                self.assertAlmostEqual(expected[name], calculated, 12, 'at index {0}\n'
                                                                       'expected:   {1:.12f}\n'
                                                                       'calculated: {2:.12f}'
                                       .format(i, expected[name], calculated))

    def testSecurityExpValueHolder(self):
        exp = SecurityExpValueHolder('close')

        expected = {}
        for i in range(len(self.aapl['close'])):
            data = dict(aapl=dict(close=self.aapl['close'][i],
                                  open=self.aapl['open'][i]),
                        ibm=dict(close=self.ibm['close'][i],
                                 open=self.ibm['open'][i]))
            exp.push(data)

            value = exp.value
            for name in value.index():
                expected[name] = np.exp(self.dataSet[name]['close'][i])
                calculated = value[name]
                self.assertAlmostEqual(expected[name], calculated, 12, 'at index {0}\n'
                                                                       'expected:   {1:.12f}\n'
                                                                       'calculated: {2:.12f}'
                                       .format(i, expected[name], calculated))

    def testSecurityLogValueHolder(self):
        logExp = SecurityLogValueHolder(SecurityExpValueHolder('close'))

        expected = {}
        for i in range(len(self.aapl['close'])):
            data = dict(aapl=dict(close=self.aapl['close'][i],
                                  open=self.aapl['open'][i]),
                        ibm=dict(close=self.ibm['close'][i],
                                 open=self.ibm['open'][i]))
            logExp.push(data)

            value = logExp.value
            for name in value.index():
                expected[name] = self.dataSet[name]['close'][i]
                calculated = value[name]
                self.assertAlmostEqual(expected[name], calculated, 12, 'at index {0}\n'
                                                                       'expected:   {1:.12f}\n'
                                                                       'calculated: {2:.12f}'
                                       .format(i, expected[name], calculated))

    def testSecurityPowValueHolder(self):
        n = 2
        pow = SecurityPowValueHolder('close', n)

        expected = {}
        for i in range(len(self.aapl['close'])):
            data = dict(aapl=dict(close=self.aapl['close'][i],
                                  open=self.aapl['open'][i]),
                        ibm=dict(close=self.ibm['close'][i],
                                 open=self.ibm['open'][i]))
            pow.push(data)

            value = pow.value
            for name in value.index():
                expected[name] = math.pow(self.dataSet[name]['close'][i], n)
                calculated = value[name]
                self.assertAlmostEqual(expected[name], calculated, 12, 'at index {0}\n'
                                                                       'expected:   {1:.12f}\n'
                                                                       'calculated: {2:.12f}'
                                       .format(i, expected[name], calculated))

    def testSecuritySqrtValueHolder(self):
        n = 2
        sqrtPow = SecuritySqrtValueHolder(SecurityPowValueHolder('close', n))

        expected = {}
        for i in range(len(self.aapl['close'])):
            data = dict(aapl=dict(close=self.aapl['close'][i],
                                  open=self.aapl['open'][i]),
                        ibm=dict(close=self.ibm['close'][i],
                                 open=self.ibm['open'][i]))
            sqrtPow.push(data)

            value = sqrtPow.value
            for name in value.index():
                expected[name] = math.sqrt(math.pow(self.dataSet[name]['close'][i], n))
                calculated = value[name]
                self.assertAlmostEqual(expected[name], calculated, 12, 'at index {0}\n'
                                                                       'expected:   {1:.12f}\n'
                                                                       'calculated: {2:.12f}'
                                       .format(i, expected[name], calculated))

    def testSecurityAbsValueHolder(self):
        absHolder = SecurityAbsValueHolder('close')

        expected = {}
        for i in range(len(self.aapl['close'])):
            data = dict(aapl=dict(close=self.aapl['close'][i],
                                  open=self.aapl['open'][i]),
                        ibm=dict(close=self.ibm['close'][i],
                                 open=self.ibm['open'][i]))
            absHolder.push(data)

            value = absHolder.value
            for name in value.index():
                expected[name] = abs(self.dataSet[name]['close'][i])
                calculated = value[name]
                self.assertAlmostEqual(expected[name], calculated, 12, 'at index {0}\n'
                                                                       'expected:   {1:.12f}\n'
                                                                       'calculated: {2:.12f}'
                                       .format(i, expected[name], calculated))

    def testSecurityMACDValueHolder(self):
        short = 5
        long = 10
        macd = SecurityMACDValueHolder(short, long, 'close')
        short_average = SecurityXAverageValueHolder(short, 'close')
        long_average = SecurityXAverageValueHolder(long, 'close')

        for i in range(len(self.aapl['close'])):
            data = dict(aapl=dict(close=self.aapl['close'][i],
                                  open=self.aapl['open'][i]),
                        ibm=dict(close=self.ibm['close'][i],
                                 open=self.ibm['open'][i]))
            macd.push(data)
            short_average.push(data)
            long_average.push(data)

            value = macd.value
            for name in value.index():
                expected = short_average.value[name] - long_average.value[name]
                calculated = value[name]
                self.assertAlmostEqual(expected, calculated, 12, 'at index {0}\n'
                                                                 'expected:   {1:.12f}\n'
                                                                 'calculat: {2:.12f}'
                                       .format(i, expected, calculated))

    def testSecurityDiffValueHolder(self):

        mv = SecurityDiffValueHolder('close')

        for i in range(len(self.aapl['close'])):
            data = dict(aapl=dict(close=self.aapl['close'][i],
                                  open=self.aapl['open'][i]),
                        ibm=dict(close=self.ibm['close'][i],
                                 open=self.ibm['open'][i]))
            mv.push(data)

            if i > 0:
                value = mv.value
                for name in value.index():
                    expected = data[name]['close'] - previous_data[name]['close']
                    calculated = value[name]
                    self.assertAlmostEqual(expected, calculated, 12, 'at index {0}\n'
                                                                     'expected:   {1:.12f}\n'
                                                                     'calculat: {2:.12f}'
                                           .format(i, expected, calculated))
            previous_data = data

    def testSecurityDiffValueHolderDeepcopy(self):
        ma = SecurityDiffValueHolder('x')

        data = dict(aapl=dict(x=1.),
                    ibm=dict(x=2.))
        data2 = dict(aapl=dict(x=2.),
                     ibm=dict(x=3.))

        ma.push(data)
        ma.push(data2)

        copied = copy.deepcopy(ma)

        for name in ma.value.index():
            self.assertAlmostEqual(ma.value[name], copied.value[name])

        for v in np.random.rand(20):
            data['aapl']['x'] = v
            data['ibm']['x'] = v + 1.
            ma.push(data)

        copied = copy.deepcopy(ma)
        for name in ma.value.index():
            self.assertAlmostEqual(ma.value[name], copied.value[name])

    def testSecurityDiffValueHolderPickle(self):
        ma = SecurityDiffValueHolder('x')

        data = dict(aapl=dict(x=1.),
                    ibm=dict(x=2.))
        data2 = dict(aapl=dict(x=2.),
                     ibm=dict(x=3.))

        ma.push(data)
        ma.push(data2)

        f = tempfile.NamedTemporaryFile('w+b', delete=False)
        pickle.dump(ma, f)
        f.close()

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            for name in ma.value.index():
                self.assertAlmostEqual(ma.value[name], pickled.value[name])
        os.unlink(f.name)

        for v in np.random.rand(20):
            data['aapl']['x'] = v
            data['ibm']['x'] = v + 1.
            ma.push(data)

        f = tempfile.NamedTemporaryFile('w+b', delete=False)
        pickle.dump(ma, f)
        f.close()

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            for name in ma.value.index():
                self.assertAlmostEqual(ma.value[name], pickled.value[name])
        os.unlink(f.name)

    def testSecurityMaximumValueHolder(self):
        mm = SecurityMaximumValueHolder('open', 'close')

        for i in range(len(self.aapl['close'])):
            data = dict(aapl=dict(close=self.aapl['close'][i],
                                  open=self.aapl['open'][i]),
                        ibm=dict(close=self.ibm['close'][i],
                                 open=self.ibm['open'][i]))
            mm.push(data)

            value = mm.value
            for name in value.index():
                expected = max(getattr(self, name)['open'][i], getattr(self, name)['close'][i])
                calculated = value[name]
                self.assertAlmostEqual(expected, calculated, 12, 'at index {0}\n'
                                                                 'expected:   {1:.12f}\n'
                                                                 'calculat: {2:.12f}'
                                       .format(i, expected, calculated))

    def testSecurityMaximumValueHolderWithValueHolder(self):
        exp1 = SecurityLatestValueHolder('open')
        exp2 = SecurityLatestValueHolder('close')
        mm = SecurityMaximumValueHolder(exp1, exp2)

        for i in range(len(self.aapl['close'])):
            data = dict(aapl=dict(close=self.aapl['close'][i],
                                  open=self.aapl['open'][i]),
                        ibm=dict(close=self.ibm['close'][i],
                                 open=self.ibm['open'][i]))
            mm.push(data)

            value = mm.value
            for name in value.index():
                expected = max(getattr(self, name)['open'][i], getattr(self, name)['close'][i])
                calculated = value[name]
                self.assertAlmostEqual(expected, calculated, 12, 'at index {0}\n'
                                                                 'expected:   {1:.12f}\n'
                                                                 'calculat: {2:.12f}'
                                       .format(i, expected, calculated))

    def testSecurityMinimumValueHolder(self):
        mm = SecurityMinimumValueHolder('open', 'close')

        for i in range(len(self.aapl['close'])):
            data = dict(aapl=dict(close=self.aapl['close'][i],
                                  open=self.aapl['open'][i]),
                        ibm=dict(close=self.ibm['close'][i],
                                 open=self.ibm['open'][i]))
            mm.push(data)

            value = mm.value
            for name in value.index():
                expected = min(getattr(self, name)['open'][i], getattr(self, name)['close'][i])
                calculated = value[name]
                self.assertAlmostEqual(expected, calculated, 12, 'at index {0}\n'
                                                                 'expected:   {1:.12f}\n'
                                                                 'calculat: {2:.12f}'
                                       .format(i, expected, calculated))

    def testSecurityMinimumValueHolderWithValueHolder(self):
        exp1 = SecurityLatestValueHolder('open')
        exp2 = SecurityLatestValueHolder('close')
        mm = SecurityMinimumValueHolder(exp1, exp2)

        for i in range(len(self.aapl['close'])):
            data = dict(aapl=dict(close=self.aapl['close'][i],
                                  open=self.aapl['open'][i]),
                        ibm=dict(close=self.ibm['close'][i],
                                 open=self.ibm['open'][i]))
            mm.push(data)

            value = mm.value
            for name in value.index():
                expected = min(getattr(self, name)['open'][i], getattr(self, name)['close'][i])
                calculated = value[name]
                self.assertAlmostEqual(expected, calculated, 12, 'at index {0}\n'
                                                                 'expected:   {1:.12f}\n'
                                                                 'calculat: {2:.12f}'
                                       .format(i, expected, calculated))

    def testSecurityNormInvValueHolder(self):
        mm1 = SecurityNormInvValueHolder('open')
        mm2 = SecurityNormInvValueHolder('open', fullAcc=True)

        for i in range(len(self.aapl['close'])):
            data = dict(aapl=dict(open=norm.cdf(self.aapl['open'][i])),
                        ibm=dict(open=norm.cdf(self.ibm['open'][i])))
            mm1.push(data)
            mm2.push(data)

            value1 = mm1.value
            value2 = mm2.value
            for name in value1.index():
                expected = norm.ppf(data[name]['open'])
                calculated = value1[name]
                self.assertAlmostEqual(expected, calculated, 6, 'at index {0}\n'
                                                                'expected:   {1:.12f}\n'
                                                                'calculat: {2:.12f}'
                                       .format(i, expected, calculated))

                calculated = value2[name]
                self.assertAlmostEqual(expected, calculated, 12, 'at index {0}\n'
                                                                 'expected:   {1:.12f}\n'
                                                                 'calculat: {2:.12f}'
                                       .format(i, expected, calculated))

    def testSecurityCeilValueHolder(self):
        mm1 = SecurityCeilValueHolder('open')

        for i in range(len(self.aapl['close'])):
            data = dict(aapl=dict(open=norm.cdf(self.aapl['open'][i])),
                        ibm=dict(open=norm.cdf(self.ibm['open'][i])))
            mm1.push(data)

            value1 = mm1.value
            for name in value1.index():
                expected = math.ceil(data[name]['open'])
                calculated = value1[name]
                self.assertAlmostEqual(expected, calculated, 6, 'at index {0}\n'
                                                                'expected:   {1:.12f}\n'
                                                                'calculat: {2:.12f}'
                                       .format(i, expected, calculated))

    def testSecurityFloorValueHolder(self):
        mm1 = SecurityFloorValueHolder('open')

        for i in range(len(self.aapl['close'])):
            data = dict(aapl=dict(open=norm.cdf(self.aapl['open'][i])),
                        ibm=dict(open=norm.cdf(self.ibm['open'][i])))
            mm1.push(data)

            value1 = mm1.value
            for name in value1.index():
                expected = math.floor(data[name]['open'])
                calculated = value1[name]
                self.assertAlmostEqual(expected, calculated, 6, 'at index {0}\n'
                                                                'expected:   {1:.12f}\n'
                                                                'calculat: {2:.12f}'
                                       .format(i, expected, calculated))

    def testSecurityRoundValueHolder(self):
        mm1 = SecurityRoundValueHolder('open')

        for i in range(len(self.aapl['close'])):
            data = dict(aapl=dict(open=norm.cdf(self.aapl['open'][i])),
                        ibm=dict(open=norm.cdf(self.ibm['open'][i])))
            mm1.push(data)

            value1 = mm1.value
            for name in value1.index():
                expected = round(data[name]['open'])
                calculated = value1[name]
                self.assertAlmostEqual(expected, calculated, 6, 'at index {0}\n'
                                                                'expected:   {1:.12f}\n'
                                                                'calculat: {2:.12f}'
                                       .format(i, expected, calculated))



# -*- coding: utf-8 -*-
u"""
Created on 2017-6-7

@author: cheng.li
"""

from math import fabs


class Converged(object):

    def __init__(self, x_tol, f_tol):

        self.x_tol = x_tol
        self.f_tol = f_tol

    def check_converge(self, root_finder, e):
        root_finder.putY(e)
        return fabs(e) < self.f_tol or root_finder.bracketWidth() < self.x_tol



# -*- coding: utf-8 -*-
u"""
Created on 2016-4-18

@author: cheng.li
"""

import os
import unittest
import numpy as np
import pandas as pd
from PyFin.POpt.Optimizer import portfolio_returns
from PyFin.POpt.Optimizer import OptTarget
from PyFin.POpt.Optimizer import portfolio_optimization


class TestOptimizer(unittest.TestCase):
    def setUp(self):
        dirName = os.path.dirname(os.path.abspath(__file__))
        filePath = os.path.join(dirName, 'data/funds_data.csv')
        self.navData = pd.read_csv(filePath, index_col=0)
        self.instruments = sorted(['XT125045.XT',
                                   'XT130295.XT',
                                   'XT090981.XT',
                                   'XT113779.XT',
                                   'XT091121.XT',
                                   'XT121986.XT',
                                   'XT1205941.XT',
                                   'XT101310.XT',
                                   'XT113932.XT',
                                   'XT090776.XT',
                                   'XT112098.XT',
                                   'XT110161.XT',
                                   'XT113453.XT',
                                   'XT125853.XT'])

    def return_cal(self, navs):
        return np.log(navs[1:] / navs[:-1])

    def create_value_dict(self, names, values):
        return {name: value for name, value in zip(names, values)}

    def testPortfolioReturnsWithUnitWeight(self):
        weights = [0.0] * len(self.navData.columns)

        for i, col in enumerate(self.navData.columns):
            tmpWeights = weights[:]
            tmpWeights[i] = 1.0
            pReturns = portfolio_returns(tmpWeights, self.navData, rebalance=False)
            pReturns_ben = self.return_cal(navs=self.navData[col].values)
            self.assertTrue(np.allclose(pReturns, pReturns_ben))

    def testPortfolioReturnsWithRandomWeights(self):
        weights = np.random.uniform(0, 1.0, len(self.navData.columns))
        weights = weights / weights.sum()

        pReturns = portfolio_returns(weights, self.navData, rebalance=False)
        portNavs = np.dot(self.navData.values, weights)
        pReturns_ben = self.return_cal(navs=portNavs)
        self.assertTrue(np.allclose(pReturns, pReturns_ben))

    def testPortfolioReturnsWithRandomWeightsWithRebalance(self):
        weights = np.random.uniform(0, 1.0, len(self.navData.columns))
        weights = weights / weights.sum()

        pReturns = portfolio_returns(weights, self.navData, rebalance=True)
        returnsTable = self.return_cal(self.navData.values)
        pReturns_ben = np.dot(returnsTable, weights)
        self.assertTrue(np.allclose(pReturns, pReturns_ben))

    def testPortfolioOptimizerWithMaximumReturn(self):
        guess = [0.1 / len(self.instruments)] * len(self.instruments)
        out, fx, its, imode, smode = portfolio_optimization(guess, self.navData, OptTarget.RETURN, multiplier=50)
        matlab_res = self.create_value_dict(self.instruments, [4.75393550839560e-19,
                                                               0,
                                                               3.43546906184303e-19,
                                                               0,
                                                               0,
                                                               8.36677837797548e-20,
                                                               0,
                                                               5.76673494924112e-19,
                                                               4.05923081267658e-19,
                                                               0,
                                                               3.12250225675825e-17,
                                                               2.41998887084620e-20,
                                                               0,
                                                               1])
        for name in matlab_res:
            self.assertAlmostEqual(matlab_res[name], out[name]['weight'], 4)

    def testPortfolioOptimizerWithMinimumVolatility(self):
        guess = [0.1 / len(self.instruments)] * len(self.instruments)
        out, fx, its, imode, smode = portfolio_optimization(guess, self.navData, OptTarget.VOL, multiplier=50)
        matlab_res = self.create_value_dict(self.instruments, [0,
                                                               0,
                                                               0,
                                                               6.99175721467524e-39,
                                                               1.87467007228981e-38,
                                                               0.00376413514611989,
                                                               0,
                                                               7.75214182640125e-38,
                                                               0.000712597692602067,
                                                               0,
                                                               0,
                                                               0.994237264821637,
                                                               0.00128600233964123,
                                                               7.05296610493373e-38])
        for name in matlab_res:
            self.assertAlmostEqual(matlab_res[name], out[name]['weight'], 4)

    def testPortfolioOptimizerWithMaximumSharp(self):
        guess = [0.1 / len(self.instruments)] * len(self.instruments)
        out, fx, its, imode, smode = portfolio_optimization(guess, self.navData, OptTarget.SHARP, multiplier=50)
        matlab_res = self.create_value_dict(self.instruments, [3.72414607721023e-39,
                                                               5.58329252509498e-40,
                                                               8.50903300813402e-40,
                                                               9.42738796971393e-40,
                                                               4.83561316181437e-40,
                                                               0.00102720032914267,
                                                               4.40810381558358e-39,
                                                               0.000145735792353755,
                                                               0.000756493258973369,
                                                               3.91725820023446e-40,
                                                               0,
                                                               0.996553125979859,
                                                               0.00117153574239032,
                                                               0.000345908897280735])
        for name in matlab_res:
            self.assertAlmostEqual(matlab_res[name], out[name]['weight'], 4)

    def testPortfolioOptimizerWithMaximumReturnWithRebalance(self):
        guess = [0.1 / len(self.instruments)] * len(self.instruments)
        out, fx, its, imode, smode = portfolio_optimization(guess, self.navData, OptTarget.RETURN, multiplier=50, rebalance=True)
        matlab_res = self.create_value_dict(self.instruments, [1.59234884538490e-17,
                                                               1.64287856671892e-18,
                                                               2.31111593326468e-33,
                                                               5.02151358940695e-17,
                                                               0,
                                                               9.41672950270667e-18,
                                                               1.64287856671892e-18,
                                                               2.82144347121750e-17,
                                                               0,
                                                               2.24595602784406e-17,
                                                               2.44504165203472e-16,
                                                               -1.68518870133883e-34,
                                                               0,
                                                               1.00000000000000])
        for name in matlab_res:
            self.assertAlmostEqual(matlab_res[name], out[name]['weight'], 4)

    def testPortfolioOptimizerWithMinimumVolatilityWithRebalance(self):
        guess = [0.1 / len(self.instruments)] * len(self.instruments)
        out, fx, its, imode, smode = portfolio_optimization(guess, self.navData, OptTarget.VOL, multiplier=50, rebalance=True)
        matlab_res = self.create_value_dict(self.instruments, [1.17549435082229e-38,
                                                               0,
                                                               0,
                                                               1.68242628961440e-37,
                                                               0,
                                                               0.00209693665998001,
                                                               0,
                                                               1.53548949576161e-37,
                                                               0.00135190283349081,
                                                               0,
                                                               6.31828213566980e-38,
                                                               0.994988759101536,
                                                               0.00156240140499358,
                                                               0])
        for name in matlab_res:
            self.assertAlmostEqual(matlab_res[name], out[name]['weight'], 4)

    def testPortfolioOptimizerWithMaximumSharpWithRebalance(self):
        guess = [0.1 / len(self.instruments)] * len(self.instruments)
        out, fx, its, imode, smode = portfolio_optimization(guess, self.navData, OptTarget.SHARP, multiplier=50, rebalance=True)
        matlab_res = self.create_value_dict(self.instruments, [2.76471048953164e-22,
                                                               7.19170303955622e-21,
                                                               2.52983701805491e-21,
                                                               1.94636169061746e-22,
                                                               0,
                                                               0,
                                                               0,
                                                               0.000165267959303058,
                                                               0.00137258304784842,
                                                               0,
                                                               2.06247008802505e-22,
                                                               0.995739327032537,
                                                               0.00134638898667400,
                                                               0.00137643297363735])
        for name in matlab_res:
            self.assertAlmostEqual(matlab_res[name], out[name]['weight'], 4)



# -*- coding: utf-8 -*-
u"""
Created on 2015-7-23

@author: cheng.li
"""

import math
import numpy as np


def norm(values, h):
    values = np.square(values)
    total = h * (np.sum(values) - 0.5 * values[0] - 0.5 * values[-1])
    return math.sqrt(total)



# -*- coding: utf-8 -*-
u"""
Created on 2017-6-11

@author: cheng.li
"""

from enum import IntEnum


class Compounding(IntEnum):

    Simple = 0
    Compounded = 1
    Continuous = 2
    SimpleThenCompounded = 3
    CompoundedThenSimple = 4



# -*- coding: utf-8 -*-
u"""
Created on 2016-12-25

@author: cheng.li
"""

import datetime as dt
import pandas as pd

sample_data = pd.DataFrame(
    data={'code': [1, 2, 1, 2, 1, 2],
          'open': [2.0, 1.0, 1.5, 3.0, 2.4, 3.5],
          'close': [1.7, 1.6, 0.9, 3.8, 1.6, 2.1]},
    index=[dt.datetime(2016, 1, 1),
           dt.datetime(2016, 1, 1),
           dt.datetime(2016, 1, 2),
           dt.datetime(2016, 1, 2),
           dt.datetime(2016, 1, 3),
           dt.datetime(2016, 1, 3)]
)

sample_data = sample_data[['code', 'open', 'close']]


if __name__ == '__main__':

    from PyFin.api import MA
    ts = MA(2, 'close')
    res = ts.transform(sample_data, name='ma_2_no_code')
    print(res)


# -*- coding: utf-8 -*-
u"""
Created on 2016-4-1

@author: cheng.li
"""

from enum import Enum
from enum import unique
import numpy as np
import pandas as pd
import scipy.optimize as opt
from PyFin.POpt.Calculators import calculate_annualized_return
from PyFin.POpt.Calculators import calculate_volatility
from PyFin.POpt.Calculators import calculate_sharp
from PyFin.POpt.Calculators import calculate_sortino


def portfolio_returns(weights, nav_table, rebalance):
    if rebalance:
        return_table = np.log(nav_table.values[1:, :] / nav_table.values[:-1, :])
        returns = np.dot(return_table, weights) / sum(weights)
    else:
        final_nav = np.dot(nav_table, weights) / sum(weights)
        returns = np.log(final_nav[1:] / final_nav[:-1])
    return returns


def utility_calculator(returns, opt_type, multiplier):
    # switch between different target types
    if opt_type == OptTarget.RETURN:
        return np.exp(calculate_annualized_return(returns, multiplier)) - 1.0
    elif opt_type == OptTarget.VOL:
        return -calculate_volatility(returns, multiplier)
    elif opt_type == OptTarget.SHARP:
        return calculate_sharp(returns, multiplier)
    elif opt_type == OptTarget.SORTINO:
        return calculate_sortino(returns, multiplier)


@unique
class OptTarget(str, Enum):
    RETURN = 'RETURN'
    VOL = 'VOL'
    SHARP = 'SHARP'
    SORTINO = 'SORTINO'


def portfolio_optimization(weights, nav_table, opt_type, multiplier, rebalance=False, lb=0., ub=1.):
    if not rebalance and opt_type == OptTarget.SORTINO:
        raise ValueError("Portfolio optimization target can't be set as "
                         "maximize sortino ratio")

    if opt_type == OptTarget.SORTINO or \
                    opt_type == OptTarget.RETURN or \
                    opt_type == OptTarget.VOL or \
                    opt_type == OptTarget.SHARP or \
            (rebalance and opt_type == OptTarget.SORTINO):
        x0 = weights

        bounds = [(lb, ub) for _ in weights]

        def eq_cond(x, *args):
            return sum(x) - 1.0

        def func(weights):
            returns = portfolio_returns(weights, nav_table, rebalance)
            return -utility_calculator(returns, opt_type, multiplier)

        out, fx, its, imode, smode = opt.fmin_slsqp(func=func,
                                                    x0=x0,
                                                    bounds=bounds,
                                                    eqcons=[eq_cond],
                                                    full_output=True,
                                                    iprint=-1,
                                                    acc=1e-12,
                                                    iter=300000)

        out = {col: weight for col, weight in zip(nav_table.columns, out)}
        return pd.DataFrame(out, index=['weight']), fx, its, imode, smode



# -*- coding: utf-8 -*-
u"""
Created on 2015-8-12

@author: cheng.li
"""

from PyFin.tests.Analysis.TechnicalAnalysis.testStatefulTechnicalAnalysers import TestStatefulTechnicalAnalysis
from PyFin.tests.Analysis.TechnicalAnalysis.testStatelessTechnicalAnalysers import TestStatelessTechnicalAnalysis

__all__ = ['TestStatefulTechnicalAnalysis',
           'TestStatelessTechnicalAnalysis']



# -*- coding: utf-8 -*-
u"""
Created on 2015-7-13

@author: cheng.li
"""

from PyFin.tests.DateUtilities.testCalendar import TestCalendar
from PyFin.tests.DateUtilities.testDate import TestDate
from PyFin.tests.DateUtilities.testPeriod import TestPeriod
from PyFin.tests.DateUtilities.testSchedule import TestSchedule
from PyFin.tests.DateUtilities.testDayCounter import TestDayCounter


__all__ = ["TestCalendar",
           "TestDate",
           "TestPeriod",
           "TestSchedule",
           "TestDayCounter"]



# -*- coding: utf-8 -*-
u"""
Created on 2015-7-15

@author: cheng.li
"""

import unittest
from PyFin.DateUtilities import Date
from PyFin.Env import Settings


class TestSettings(unittest.TestCase):
    def testEvaluationDate(self):
        referenceDate = Date(2015, 1, 1)
        Settings.evaluationDate = referenceDate
        self.assertEqual(Settings.evaluationDate, referenceDate, "Evaluation Date should be \n"
                                                                 "expected:   {0}\n"
                                                                 "calculated: {1}".format(referenceDate,
                                                                                          Settings.evaluationDate))

        # check wrong input
        with self.assertRaises(ValueError):
            Settings.evaluationDate = 2



# -*- coding: utf-8 -*-
u"""
Created on 2017-1-7

@author: cheng.li
"""

import numpy as np
from scipy.optimize import least_squares
from PyFin.PricingEngines.SVIInterpolationImpl import sviVolatility
from PyFin.PricingEngines.SVIInterpolationImpl import sviVolatilities
from PyFin.PricingEngines.SVIInterpolationImpl import _sviCalibrationIteration
from PyFin.PricingEngines.SVIInterpolationImpl import _parametersCheck


def sviCalibration(strikes,
                   volatilites,
                   forward,
                   expiryTime,
                   initialA,
                   initialB,
                   initialSigma,
                   initialRho,
                   initialM,
                   isFixedA=False,
                   isFixedB=False,
                   isFixedSigma=False,
                   isFixedRho=False,
                   isFixedM=False,
                   method='trf'):
    x0, freeParameters, fixedParameters, bounds = _parametersCheck(initialA,
                                                                   initialB,
                                                                   initialSigma,
                                                                   initialRho,
                                                                   initialM,
                                                                   isFixedA,
                                                                   isFixedB,
                                                                   isFixedSigma,
                                                                   isFixedRho,
                                                                   isFixedM)

    if method != 'lm':
        x = least_squares(_sviCalibrationIteration,
                          x0,
                          method=method,
                          bounds=bounds,
                          ftol=1e-10,
                          gtol=1e-10,
                          xtol=1e-10,
                          args=(freeParameters, strikes, volatilites, forward, expiryTime, fixedParameters))
    else:
        x = least_squares(_sviCalibrationIteration,
                          x0,
                          method=method,
                          ftol=1e-10,
                          gtol=1e-10,
                          xtol=1e-10,
                          args=(freeParameters, strikes, volatilites, forward, expiryTime, fixedParameters))

    parameters = ['a', 'b', 'sigma', 'rho', 'm']
    calibratedParameters = dict(zip(freeParameters, x.x))

    res = []
    for name in parameters:
        try:
            res.append(calibratedParameters[name])
        except KeyError:
            res.append(fixedParameters[name])

    return np.array(res), x.status, x.message


__all__ = ['sviVolatility',
           'sviVolatilities',
           'sviCalibration']


# -*- coding: utf-8 -*-
u"""
Created on 2015-7-23

@author: cheng.li
"""

from PyFin.PricingEngines.BlackFormula import bachelierFormula
from PyFin.PricingEngines.BlackFormula import bachelierFormulaImpliedVol
from PyFin.PricingEngines.BlackFormula import blackFormula
from PyFin.PricingEngines.BlackFormula import blackFormula2
from PyFin.PricingEngines.BlackFormula import blackFormulaImpliedStdDev
from PyFin.PricingEngines.BlackFormula import blackFormulaImpliedVol
from PyFin.PricingEngines.SabrFormula import sabrVolatility
from PyFin.PricingEngines.SabrFormula import sabrVolatilities
from PyFin.PricingEngines.SabrFormula import sabrCalibration
from PyFin.PricingEngines.SVIInterpolation import sviVolatility
from PyFin.PricingEngines.SVIInterpolation import sviVolatilities
from PyFin.PricingEngines.SVIInterpolation import sviCalibration


__all__ = ['bachelierFormula',
           'bachelierFormulaImpliedVol',
           'blackFormula',
           'blackFormula2',
           'blackFormulaImpliedStdDev',
           'blackFormulaImpliedVol',
           'sabrVolatility',
           'sabrVolatilities',
           'sabrCalibration',
           'sviVolatility',
           'sviVolatilities',
           'sviCalibration']



# -*- coding: utf-8 -*-
u"""
Created on 2016-12-27

@author: cheng.li
"""

import datetime as dt
import numpy as np
import pandas as pd
from PyFin.api import *
from PyFin.Math.Accumulators import MovingAverage

n = 300
m = 60

index = pd.date_range(dt.datetime(2021, 1, 1), dt.datetime(2021, 1, 1) + dt.timedelta(days=m-1))
index = np.repeat(index, n)

df = pd.DataFrame(np.random.randn(n*m, 3), columns=['x', 'y', 'z'], index=index)
df['c'] = np.concatenate([np.linspace(0, n-1, n, dtype=int)] * m)

start = dt.datetime.now()
t = MA(20, 'x') / MA(30, 'y')
res = t.transform(df, category_field='c')
print("Finance-Python (group ma): {0}s".format(dt.datetime.now() - start))

start = dt.datetime.now()
groups = df.groupby('c')
res = groups['x'].rolling(20).mean() / groups['y'].rolling(30).mean()
print("Pandas (group ma): {0}s".format(dt.datetime.now() - start))

start = dt.datetime.now()
t = MovingAverage(20, 'x') / MovingAverage(30, 'x')
res = t.transform(df)
print("\nFinance-Python (rolling ma): {0}s".format(dt.datetime.now() - start))

start = dt.datetime.now()
res = df['x'].rolling(20).mean() / df['x'].rolling(30).mean()
print("Pandas (rolling ma): {0}s".format(dt.datetime.now() - start))


"""
Cross section analysis examples ...
"""

# rank

index = pd.date_range(dt.datetime(1990, 1, 1), dt.datetime(1990, 1, 1) + dt.timedelta(days=m-1))
index = np.repeat(index, n)

df = pd.DataFrame(np.random.randn(n*m, 1), columns=['x'], index=index)
ind = np.random.randint(0, int(n / 100), len(df))
df['c'] = np.concatenate([np.linspace(0, n-1, n, dtype=int)] * m)
df['ind'] = ind

start = dt.datetime.now()
t = CSRank('x', groups='ind')
res1 = t.transform(df, category_field='c')
print("\nFinance-Python (cs rank): {0}s".format(dt.datetime.now() - start))

start = dt.datetime.now()
df2 = df.reset_index()
res2 = df2.groupby(['index', 'ind']).apply(lambda x: x['x'].rank())
print("Pandas (cs rank): {0}s".format(dt.datetime.now() - start))

res2 = pd.DataFrame({'index': res2.index.get_level_values(2), 'exp_rank': res2.values})
res2.sort_values('index', inplace=True)
res1['exp_rank'] = res2['exp_rank'].values
diff = res1['transformed'] - res1['exp_rank']
print("total rank difference: {0}".format(np.abs(diff).sum()))

# percentile

start = dt.datetime.now()
t = CSQuantiles('x', groups='ind')
res1 = t.transform(df, category_field='c')
print("\nFinance-Python (cs percentile): {0}s".format(dt.datetime.now() - start))

start = dt.datetime.now()
df2 = df.reset_index()
res2 = df2.groupby(['index', 'ind']).apply(lambda x: x['x'].rank() / (len(x) + 1))
print("Pandas (cs percentile): {0}s".format(dt.datetime.now() - start))

res2 = pd.DataFrame({'index': res2.index.get_level_values(2), 'exp_rank': res2.values})
res2.sort_values('index', inplace=True)
res1['exp_rank'] = res2['exp_rank'].values
diff = res1['transformed'] - res1['exp_rank']
print("total percentile difference: {0}".format(np.abs(diff).sum()))

# zscore

start = dt.datetime.now()
t = CSZScore('x', groups='ind')
res1 = t.transform(df, category_field='c')
print("\nFinance-Python (cs zscore): {0}s".format(dt.datetime.now() - start))

start = dt.datetime.now()
df2 = df.reset_index()
res2 = df2.groupby(['index', 'ind']).apply(lambda x: (x['x'] - x['x'].mean()) / x['x'].std(ddof=0))
print("Pandas (cs zscore): {0}s".format(dt.datetime.now() - start))

res2 = pd.DataFrame({'index': res2.index.get_level_values(2), 'exp_rank': res2.values})
res2.sort_values('index', inplace=True)
res1['exp_rank'] = res2['exp_rank'].values
diff = res1['transformed'] - res1['exp_rank']
print("total zscore difference: {0}".format(np.abs(diff).sum()))


# -*- coding: utf-8 -*-
u"""
Created on 2017-1-4

@author: cheng.li
"""

import unittest
import copy
import tempfile
import pickle
import os
from collections import deque
import numpy as np
from PyFin.Math.Accumulators.impl import (
    Deque,
    DiffDeque,
    UniqueDiffDeque
)


class TestAccumulatorImpl(unittest.TestCase):

    def testDequeDeepCopy(self):
        benchmark_deque = Deque(5)

        copied_deque = copy.deepcopy(benchmark_deque)

        self.assertEqual(copied_deque, benchmark_deque)

    def testDequePickle(self):
        benchmark_deque = Deque(5)

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(benchmark_deque, f)

        with open(f.name, 'rb') as f2:
            pickled_deque = pickle.load(f2)
            self.assertEqual(benchmark_deque, pickled_deque)

        os.unlink(f.name)

    def testDequeSum(self):
        deque = Deque(5)
        deque.dumps([1., 2., 3., 4., 5., 6.])
        self.assertEqual(deque.sum(), 20.)

    def testDequeIdx(self):
        deque1 = deque(maxlen=10)
        deque2 = Deque(10)

        values = np.random.randn(1000)
        for v in values:
            deque1.append(v)
            deque2.dumps([v])

            max_val = max(deque1)
            expected_idx = np.where(np.array(deque1) == max_val)[0][0]
            caclulated_idx = deque2.idx(max_val)
            self.assertEqual(expected_idx, caclulated_idx)

    def testDiffDequeDeepCopy(self):
        benchmark_deque = DiffDeque(5, closed="left")
        copied_deque = copy.deepcopy(benchmark_deque)
        self.assertEqual(copied_deque, benchmark_deque)

    def testDiffDequeSum(self):
        deque = DiffDeque(10)
        values = [1, 2, 3, 4, 5]
        stamps = [1, 7, 9, 13, 26]

        deque.dumps(values, stamps)
        self.assertEqual(deque.sum(), 5)

    def testDiffDequeWithLeftClosedSum(self):
        deque = DiffDeque(10, closed="left")

        values = [1, 2, 3, 4, 5]
        stamps = [1, 7, 9, 16, 26]

        deque.dumps(values, stamps)
        self.assertEqual(deque.sum(), 4)

    def testDiffDequeWithBothClosedSum(self):
        deque = DiffDeque(10, closed="both")

        values = [1, 2, 3, 4, 5]
        stamps = [1, 7, 9, 16, 26]

        deque.dumps(values, stamps)
        self.assertEqual(deque.sum(), 9)

    def testDiffDequeWithNeitherClosedSum(self):
        deque = DiffDeque(10, closed="neither")

        values = [1, 2, 3, 4, 5]
        stamps = [1, 7, 9, 16, 26]

        deque.dumps(values, stamps)
        self.assertEqual(deque.sum(), 0)

    def testUniqueDiffDequeDeepCopy(self):
        benchmark_deque = UniqueDiffDeque(5, closed="left")
        copied_deque = copy.deepcopy(benchmark_deque)
        self.assertEqual(copied_deque, benchmark_deque)

    def testUniqueDiffDequeSum(self):
        deque = UniqueDiffDeque(10)
        values = [2, 2, 3, 5, 5]
        stamps = [1, 7, 20, 22, 26]

        deque.dumps(values, stamps)
        self.assertEqual(deque.sum(), 8)

    def testUniqueDiffDequeWithLeftClosedSum(self):
        deque = UniqueDiffDeque(10, closed="left")

        values = [2, 2, 5, 5, 5]
        stamps = [1, 16, 20, 22, 26]

        deque.dumps(values, stamps)
        self.assertEqual(deque.sum(), 7)

    def testUniqueDiffDequeWithBothClosedSum(self):
        deque = UniqueDiffDeque(10, closed="both")

        values = [2, 2, 5, 5, 7]
        stamps = [1, 16, 20, 22, 26]

        deque.dumps(values, stamps)
        self.assertEqual(deque.sum(), 14)

    def testUniqueDiffDequeWithNeitherClosedSum(self):
        deque = UniqueDiffDeque(10, closed="neither")

        values = [2, 2, 5, 5, 7]
        stamps = [1, 16, 20, 22, 26]

        deque.dumps(values, stamps)
        self.assertEqual(deque.sum(), 5)







# -*- coding: utf-8 -*-
u"""
Created on 2015-7-15

@author: cheng.li
"""

import unittest
import copy
import tempfile
import pickle
import os
from PyFin.DateUtilities import Date
from PyFin.DateUtilities import Schedule
from PyFin.DateUtilities import Period
from PyFin.DateUtilities import Calendar
from PyFin.Enums import TimeUnits
from PyFin.Enums import BizDayConventions


class TestSchedule(unittest.TestCase):
    def checkDates(self, s, expected):
        if s.size() != len(expected):
            self.fail("expected {0:d} dates, found {1}".format(len(expected), s.size()))

        for i in range(s.size()):
            if s[i] != expected[i]:
                self.fail("expected {0} at index found {1}".format(expected[i], s[i]))

    def testScheduleInitialize(self):
        startDate = Date(2013, 3, 31)
        endDate = Date(2013, 6, 30)
        tenor = Period('1m')
        cal = Calendar('NullCalendar')
        sch = Schedule(startDate, endDate, tenor, cal)
        expected = [Date(2013, 3, 31), Date(2013, 4, 30), Date(2013, 5, 31), Date(2013, 6, 30)]
        for i in range(sch.size()):
            self.assertEqual(expected[i], sch[i])

    def testScheduleInitializeWithYearly(self):
        startDate = Date(2012, 2, 29)
        endDate = Date(2013, 3, 1)
        tenor = Period('1y')
        cal = Calendar('NullCalendar')
        sch = Schedule(startDate, endDate, tenor, cal)
        expected = [Date(2012, 2, 29), Date(2013, 2, 28), Date(2013, 3, 1)]
        for i in range(sch.size()):
            self.assertEqual(expected[i], sch[i])

    def testDailySchedule(self):
        # Jan 2 and Jan 3 are skipped as New Year holiday
        # Jan 7 is skipped as weekend
        # Jan 8 is adjusted to Jan 9 with following convention
        startDate = Date(2012, 1, 1)
        s = Schedule(startDate,
                     startDate + 7,
                     Period(length=1, units=TimeUnits.Days),
                     Calendar("China.SSE"),
                     BizDayConventions.Preceding)

        expected = [Date(2011, 12, 30), Date(2012, 1, 4), Date(2012, 1, 5), Date(2012, 1, 6), Date(2012, 1, 9)]
        self.checkDates(s, expected)

        # The schedule should skip Saturday 21st and Sunday 22rd.
        # Previously, it would adjust them to Friday 20th, resulting
        # in three copies of the same date.
        startDate = Date(2012, 1, 17)
        s = Schedule(startDate,
                     startDate + 7,
                     Period(length=1, units=TimeUnits.Days),
                     Calendar("Target"),
                     BizDayConventions.Preceding)
        expected = [Date(2012, 1, 17), Date(2012, 1, 18), Date(2012, 1, 19), Date(2012, 1, 20), Date(2012, 1, 23),
                    Date(2012, 1, 24)]
        self.checkDates(s, expected)

    def testScheduleDeepCopy(self):
        startDate = Date(2013, 3, 31)
        endDate = Date(2013, 6, 30)
        tenor = Period('1m')
        cal = Calendar('NullCalendar')
        sch = Schedule(startDate, endDate, tenor, cal)
        copied_sch = copy.deepcopy(sch)

        self.assertEqual(sch, copied_sch)

    def testSchedulePickle(self):
        startDate = Date(2013, 3, 31)
        endDate = Date(2013, 6, 30)
        tenor = Period('1m')
        cal = Calendar('NullCalendar')
        sch = Schedule(startDate, endDate, tenor, cal)

        f = tempfile.NamedTemporaryFile('w+b', delete=False)
        pickle.dump(sch, f)
        f.close()

        with open(f.name, 'rb') as f2:
            pickled_sch = pickle.load(f2)
            self.assertEqual(sch, pickled_sch)

        os.unlink(f.name)



# -*- coding: utf-8 -*-
u"""
Created on 2016-4-1

@author: cheng.li
"""

import math
from PyFin.Math.Accumulators import MovingSharp
from PyFin.Math.Accumulators import MovingSortino


def calculate_annualized_return(returns, multplier=50):
    return returns.mean() * multplier


def calculate_volatility(returns, multplier=50):
    return returns.std(ddof=1) * math.sqrt(multplier)


def calculate_sharp(returns, multplier=50):
    msharp = MovingSharp(len(returns), x='ret', y='riskFree')
    for ret in returns:
        msharp.push({'ret': ret, 'riskFree': 0.})
    return msharp.value * math.sqrt(multplier)


def calculate_sortino(returns, multplier=50):
    msortino = MovingSortino(len(returns))
    for ret in returns:
        msortino.push({'ret': ret, 'riskFree': 0.})
    return msortino.value * math.sqrt(multplier)



# -*- coding: utf-8 -*-
u"""
Created on 2017-1-7

@author: cheng.li
"""

import unittest
import numpy as np
from PyFin.PricingEngines import sviVolatilities
from PyFin.PricingEngines import sviCalibration


class TestSVIInterpolation(unittest.TestCase):
    def setUp(self):
        self.strikes = np.array([0.] * 31)
        self.expiry = 1.0
        self.forward = 0.039

        self.strikes[0] = 0.03
        self.strikes[1] = 0.032
        self.strikes[2] = 0.034
        self.strikes[3] = 0.036
        self.strikes[4] = 0.038
        self.strikes[5] = 0.04
        self.strikes[6] = 0.042
        self.strikes[7] = 0.044
        self.strikes[8] = 0.046
        self.strikes[9] = 0.048
        self.strikes[10] = 0.05
        self.strikes[11] = 0.052
        self.strikes[12] = 0.054
        self.strikes[13] = 0.056
        self.strikes[14] = 0.058
        self.strikes[15] = 0.06
        self.strikes[16] = 0.062
        self.strikes[17] = 0.064
        self.strikes[18] = 0.066
        self.strikes[19] = 0.068
        self.strikes[20] = 0.07
        self.strikes[21] = 0.072
        self.strikes[22] = 0.074
        self.strikes[23] = 0.076
        self.strikes[24] = 0.078
        self.strikes[25] = 0.08
        self.strikes[26] = 0.082
        self.strikes[27] = 0.084
        self.strikes[28] = 0.086
        self.strikes[29] = 0.088
        self.strikes[30] = 0.09

        self.initialA = -0.0410
        self.initialB = 0.1331
        self.initialSigma = 0.4153
        self.initialRho = 0.3060
        self.initialM = 0.3586

    def testSVICalibration(self):
        expectedVolatilities = sviVolatilities(self.strikes,
                                               self.forward,
                                               self.expiry,
                                               self.initialA,
                                               self.initialB,
                                               self.initialSigma,
                                               self.initialRho,
                                               self.initialM)

        x = sviCalibration(self.strikes, expectedVolatilities, self.forward, self.expiry, 0.01, 0.01, 0.01, 0.01, 0.01)
        calibratedParameters = x[0]
        calculatedVolatilities = sviVolatilities(self.strikes,
                                                 self.forward,
                                                 self.expiry,
                                                 calibratedParameters[0],
                                                 calibratedParameters[1],
                                                 calibratedParameters[2],
                                                 calibratedParameters[3],
                                                 calibratedParameters[4])

        np.testing.assert_array_almost_equal(expectedVolatilities, calculatedVolatilities)



# -*- coding: utf-8 -*-
u"""
Created on 2015-7-23

@author: cheng.li
"""

from PyFin.tests.Math import Distributions
from PyFin.tests.Math import Accumulators
from PyFin.tests.Math import Timeseries
from PyFin.tests.Math import RootFinder

__all__ = ["Distributions",
           "Accumulators",
           "Timeseries",
           "RootFinder"]



# -*- coding: utf-8 -*-
u"""
Created on 2015-7-13

@author: cheng.li
"""

from PyFin.DateUtilities import Date
from PyFin.DateUtilities import Calendar
from PyFin.DateUtilities import Period
from PyFin.DateUtilities import Schedule
from PyFin.Enums import BizDayConventions
from PyFin.Enums import DateGeneration
from PyFin.Enums import TimeUnits
from PyFin.DateUtilities import check_date
from PyFin.DateUtilities import check_period


def isBizDay(holidayCenter, ref):
    cal = Calendar(holidayCenter)
    ref = check_date(ref)
    return cal.isBizDay(ref)


def datesList(fromDate, toDate):
    fromDate = check_date(fromDate)
    toDate = check_date(toDate)
    return [Date.fromExcelSerialNumber(serial).toDateTime() for serial in
            range(fromDate.serialNumber, toDate.serialNumber + 1)]


def bizDatesList(holidayCenter, fromDate, toDate):
    cal = Calendar(holidayCenter)
    fromDate = check_date(fromDate)
    toDate = check_date(toDate)
    return [d.toDateTime() for d in cal.bizDatesList(fromDate, toDate)]


def holDatesList(holidayCenter, fromDate, toDate, includeWeekend=True):
    cal = Calendar(holidayCenter)
    fromDate = check_date(fromDate)
    toDate = check_date(toDate)
    return [d.toDateTime() for d in cal.holDatesList(fromDate, toDate, includeWeekend)]


def advanceDate(referenceDate, period):
    d = check_date(referenceDate) + period
    return d.toDateTime()


def adjustDateByCalendar(holidayCenter, referenceDate, convention=BizDayConventions.Following):
    cal = Calendar(holidayCenter)
    refer = check_date(referenceDate)
    return cal.adjustDate(refer, convention).toDateTime()


def advanceDateByCalendar(holidayCenter, referenceDate, period, convention=BizDayConventions.Following):
    cal = Calendar(holidayCenter)
    refer = check_date(referenceDate)
    period = check_period(period)
    return cal.advanceDate(refer, period, convention).toDateTime()


def nthWeekDay(nth, dayOfWeek, month, year):
    date = Date.nthWeekday(nth, dayOfWeek, month, year)
    return date.toDateTime()


def makeSchedule(firstDate,
                 endDate,
                 tenor,
                 calendar='NullCalendar',
                 dateRule=BizDayConventions.Following,
                 dateGenerationRule=DateGeneration.Forward):

    cal = Calendar(calendar)
    firstDate = check_date(firstDate)
    endDate = check_date(endDate)
    tenor = check_period(tenor)

    if tenor.units() == TimeUnits.BDays:
        schedule = []
        if dateGenerationRule == DateGeneration.Forward:
            d = cal.adjustDate(firstDate, dateRule)
            while d <= endDate:
                schedule.append(d)
                d = cal.advanceDate(d, tenor, dateRule)
        elif dateGenerationRule == DateGeneration.Backward:
            d = cal.adjustDate(endDate, dateRule)
            while d >= firstDate:
                schedule.append(d)
                d = cal.advanceDate(d, -tenor, dateRule)
            schedule = sorted(schedule)
    else:
        schedule = Schedule(firstDate, endDate, tenor, cal, convention=dateRule, dateGenerationRule=dateGenerationRule)
    return [d.toDateTime() for d in schedule]



# -*- coding: utf-8 -*-
u"""
Created on 2015-7-23

@author: cheng.li
"""

import unittest
import math
from PyFin.Enums.OptionType import OptionType
from PyFin.PricingEngines.BlackFormula import bachelierFormula
from PyFin.PricingEngines.BlackFormula import bachelierFormulaImpliedVol
from PyFin.PricingEngines.BlackFormula import blackFormula
from PyFin.PricingEngines.BlackFormula import blackFormula2
from PyFin.PricingEngines.BlackFormula import blackFormulaImpliedStdDev
from PyFin.PricingEngines.BlackFormula import blackFormulaImpliedVol


class TestBlackFormula(unittest.TestCase):

    def testBlackBasic(self):
        strike = 1.0
        forward = 1.2
        stdDev = 0.2
        discount = 0.95

        expected = discount * (forward - strike)
        calculated = blackFormula(OptionType.Call, strike, forward, 0.0, discount)
        self.assertAlmostEqual(expected, calculated, 15)

        expected = discount * forward
        calculated = blackFormula(OptionType.Call, 0.0, forward, stdDev, discount)
        self.assertAlmostEqual(expected, calculated, 15)

    def testBachelierBasics(self):
        strike = 1.0
        forward = 1.2
        discount = 0.95

        expected = discount * (forward - strike)
        calculated = bachelierFormula(OptionType.Call, strike, forward, 0.0, discount)
        self.assertAlmostEqual(expected, calculated, 15)

    def testBlackImpliedStdDev(self):
        forward = 1.0
        bpvol = 0.25
        tte = 10.0
        stdDev = bpvol * math.sqrt(tte)
        discount = math.exp(-0.05 * tte)

        dList = [-3.0, -2.0, -1.0, -0.5, 0.0, 0.5, 1.0, 2.0, 3.0]

        for d in dList:
            strike = forward * math.exp(d * bpvol * math.sqrt(tte))
            callPrem = blackFormula(OptionType.Call, strike, forward, stdDev, discount)
            impliedStdDev = blackFormulaImpliedStdDev(OptionType.Call, strike, forward, callPrem, discount)
            self.assertAlmostEqual(impliedStdDev, stdDev, 8)

            putPrem = blackFormula(OptionType.Put, strike, forward, stdDev, discount)
            impliedStdDev = blackFormulaImpliedStdDev(OptionType.Put, strike, forward, putPrem, discount)
            self.assertAlmostEqual(impliedStdDev, stdDev, 8)

    def testBlackImpliedVol(self):
        forward = 1.0
        bpvol = 0.25
        tte = 10.0
        riskFree = 0.05

        dList = [-3.0, -2.0, -1.0, -0.5, 0.0, 0.5, 1.0, 2.0, 3.0]

        for d in dList:
            strike = forward * math.exp(d * bpvol * math.sqrt(tte))
            callPrem = blackFormula2(OptionType.Call, strike, forward, tte, bpvol, riskFree)
            impliedVol = blackFormulaImpliedVol(OptionType.Call, strike, forward, tte, callPrem, riskFree)
            self.assertAlmostEqual(impliedVol, bpvol, 8)

            putPrem = blackFormula2(OptionType.Put, strike, forward, tte, bpvol, riskFree)
            impliedVol = blackFormulaImpliedVol(OptionType.Put, strike, forward, tte, putPrem, riskFree)
            self.assertAlmostEqual(impliedVol, bpvol, 8)

    def testBachelierImpliedVol(self):
        forward = 1.0
        bpvol = 0.01
        tte = 10.0
        r = 0.03

        dList = [-3.0, -2.0, -1.0, -0.5, 0.0, 0.5, 1.0, 2.0, 3.0]

        for d in dList:
            strike = forward + d * bpvol * math.sqrt(tte)
            callPrem = bachelierFormula(OptionType.Call, strike, forward, bpvol * math.sqrt(tte), math.exp(-r * tte))
            impliedBpVol = bachelierFormulaImpliedVol(OptionType.Call, strike, forward, tte, callPrem,
                                                      math.exp(-r * tte))
            self.assertAlmostEqual(impliedBpVol, bpvol, 12)

            callPrem = bachelierFormula(OptionType.Put, strike, forward, bpvol * math.sqrt(tte), math.exp(-r * tte))
            impliedBpVol = bachelierFormulaImpliedVol(OptionType.Put, strike, forward, tte, callPrem,
                                                      math.exp(-r * tte))
            self.assertAlmostEqual(impliedBpVol, bpvol, 12)



# -*- coding: utf-8 -*-
u"""
Created on 2015-7-9

@author: cheng.li
"""

from PyFin.DateUtilities.Calendar import Calendar
from PyFin.DateUtilities.Date import Date
from PyFin.DateUtilities.Date import check_date
from PyFin.DateUtilities.Period import Period
from PyFin.DateUtilities.Period import check_period
from PyFin.DateUtilities.Schedule import Schedule

__all__ = ['Calendar',
           'Date',
           'Period',
           'Schedule',
           'check_date',
           'check_period']



# -*- coding: utf-8 -*-
u"""
Created on 2017-1-7

@author: cheng.li
"""

import numpy as np
from scipy.optimize import least_squares
from PyFin.PricingEngines.SabrFormulaImpl import sabrVolatility
from PyFin.PricingEngines.SabrFormulaImpl import sabrVolatilities
from PyFin.PricingEngines.SabrFormulaImpl import _sabrCalibrationIteration
from PyFin.PricingEngines.SabrFormulaImpl import _parametersCheck


def sabrCalibration(strikes,
                    volatilites,
                    forward,
                    expiryTime,
                    intialAlpha,
                    initialBeta,
                    initialNu,
                    initialRho,
                    isFixedAlpha=False,
                    isFixedBeta=False,
                    isFixedNu=False,
                    isFixedRho=False,
                    method='trf'):

    x0, freeParameters, fixedParameters, bounds = _parametersCheck(intialAlpha,
                                                                   initialBeta,
                                                                   initialNu,
                                                                   initialRho,
                                                                   isFixedAlpha,
                                                                   isFixedBeta,
                                                                   isFixedNu,
                                                                   isFixedRho)

    if method != 'lm':
        x = least_squares(_sabrCalibrationIteration,
                          x0,
                          method=method,
                          bounds=bounds,
                          ftol=1e-10,
                          gtol=1e-10,
                          xtol=1e-10,
                          args=(freeParameters, strikes, volatilites, forward, expiryTime, fixedParameters))
    else:
        x = least_squares(_sabrCalibrationIteration,
                          x0,
                          method=method,
                          ftol=1e-10,
                          gtol=1e-10,
                          xtol=1e-10,
                          args=(freeParameters, strikes, volatilites, forward, expiryTime, fixedParameters))

    parameters = ['alpha', 'beta', 'nu', 'rho']
    calibratedParameters = dict(zip(freeParameters, x.x))

    res = []
    for name in parameters:
        try:
            res.append(calibratedParameters[name])
        except KeyError:
            res.append(fixedParameters[name])

    return np.array(res), x.status, x.message


__all__ = ['sabrVolatility',
           'sabrVolatilities',
           'sabrCalibration']


# -*- coding: utf-8 -*-
u"""
Created on 2015-7-21

@author: cheng.li
"""

from PyFin.Enums.NormalizingType import NormalizingType
from PyFin.DateUtilities.Calendar import Calendar


class Normalizer(object):
    def __init__(self, type=NormalizingType.Null):
        self._previous = None
        self._current = None
        self._type = type

        if self._type == NormalizingType.BizDay or self._type == NormalizingType.CalendarDay:
            self._isFirst = True
            if self._type == NormalizingType.BizDay:
                self._cal = Calendar('China.SSE')

    def normalizeOneDayReturn(self, pDate, pReturn):
        if self._type == NormalizingType.Null:
            return pReturn
        elif self._type == NormalizingType.CalendarDay:
            return self._calendarDayCalcualtion(pDate, pReturn)
        elif self._type == NormalizingType.BizDay:
            return self._bizDayCalculation(pDate, pReturn)

    def _calendarDayCalcualtion(self, pDate, pReturn):
        if self._isFirst:
            self._isFirst = False
        else:
            daysBetween = pDate - self._previous
            pReturn /= daysBetween

        self._previous = pDate
        return pReturn

    def _bizDayCalculation(self, pDate, pReturn):
        if self._isFirst:
            self._isFirst = False
        else:
            daysBetween = self._cal.bizDaysBetween(self._previous, pDate, True, False)
            pReturn /= daysBetween

        self._previous = pDate
        return pReturn



# -*- coding: utf-8 -*-
u"""
Created on 2017-6-12

@author: cheng.li
"""

from PyFin.CashFlows.CashFlow import Coupon
from PyFin.CashFlows.InterestRate import InterestRate
from PyFin.Enums.Compoundings import Compounding
from PyFin.Enums.Frequencies import Frequency
from PyFin.Enums.BizDayConventions import BizDayConventions


class FixedRateCoupon(Coupon):

    def __init__(self,
                 paymentDate,
                 norminal,
                 interestRate,
                 accrualStartDate,
                 accrualEndDate,
                 refPeriodStart=None,
                 refPeriodEnd=None,
                 exCouponDate=None):
        super(FixedRateCoupon, self).__init__(paymentDate,
                                              norminal,
                                              accrualStartDate,
                                              accrualEndDate,
                                              refPeriodStart,
                                              refPeriodEnd,
                                              exCouponDate)
        self.rate_ = interestRate

    def amount(self):
        return self.norminal_ * (self.rate_.compoundFactor(self.accrualStartDate_,
                                                           self.accrualEndDate_,
                                                           self.refPeriodStart_,
                                                           self.refPeriodEnd_) - 1.0)

    def accruedAmount(self, d):
        if d <= self.accrualStartDate_ or d > self.paymentDate_:
            return 0.
        elif self.tradingExCoupon(d):
            return -self.norminal_ * (self.rate_.compoundFactor(d,
                                                                self.accrualEndDate_,
                                                                self.refPeriodStart_,
                                                                self.refPeriodEnd_) - 1.0)
        else:
            return self.norminal_ * (self.rate_.compoundFactor(self.accrualStartDate_,
                                                               min(d, self.accrualEndDate_),
                                                               self.refPeriodStart_,
                                                               self.refPeriodEnd_) - 1.0)


class FixedRateLeg(object):

    def __init__(self, schedule):
        self.schedule_ = schedule
        self.calendar_ = self.schedule_.calendar()
        self.paymentAdjustment_ = BizDayConventions.Following
        self.firstPeriodDC_ = None
        self.notionals_ = None
        self.couponRates_ = None
        self.exCouponPeriod_ = None
        self.exCouponCalendar_ = None
        self.exCouponAdjustment_ = None
        self.exCouponEndOfMonth_ = None

    def withNotionals(self, notionals):
        self.notionals_ = notionals
        return self

    def withCouponRate(self, rate, dc, comp, freq):
        self.couponRates_ = []
        self.couponRates_.append(InterestRate(rate, dc, comp, freq))
        return self

    def withCouponRateByInterestRate(self, ir):
        self.couponRates_ = []
        self.couponRates_.append(ir)
        return self

    def withPaymentAdjustment(self, convention):
        self.paymentAdjustment_ = convention
        return self

    def withFirstPeriodDayCounter(self, dc):
        self.firstPeriodDC_ = dc
        return self

    def withPaymentCalendar(self, cal):
        self.calendar_ = cal
        return self

    def withExCouponPeriod(self, period, cal, convention, endOfMonth):
        self.exCouponPeriod_ = period
        self.exCouponCalendar_ = cal
        self.exCouponAdjustment_ = convention
        self.exCouponEndOfMonth_ = endOfMonth
        return self

    def toLeg(self):
        leg = []
        start = self.schedule_._dates[0]
        end = self.schedule_._dates[1]
        paymentDate = self.calendar_.adjustDate(end, self.paymentAdjustment_)
        rate = self.couponRates_[0]
        nominal = self.notionals_[0]

        exCouponDate = None

        if self.exCouponPeriod_:
            exCouponDate = self.exCouponCalendar_.advanceDate(paymentDate,
                                                              -self.exCouponPeriod_,
                                                              self.exCouponAdjustment_,
                                                              self.exCouponEndOfMonth_)

        if self.schedule_.isRegular(1):
            leg.append(FixedRateCoupon(paymentDate, nominal, rate, start, end, start, end, exCouponDate))
        else:
            ref = self.schedule_.calendar().advanceDate(end,
                                                        -self.schedule_.tenor(),
                                                        self.schedule_._convention,
                                                        self.schedule_.endOfMonth())
            r = InterestRate(rate.r_,
                             self.firstPeriodDC_ if self.firstPeriodDC_ else rate.dayCounter_,
                             rate.comp_,
                             rate.freq_)
            leg.append(FixedRateCoupon(paymentDate,
                                       nominal,
                                       r,
                                       start,
                                       end,
                                       ref,
                                       end,
                                       exCouponDate))

        for i in range(2, self.schedule_.size()-1):
            start = end
            end = self.schedule_._dates[i]
            paymentDate = self.calendar_.adjustDate(end, self.paymentAdjustment_)

            if self.exCouponPeriod_:
                exCouponDate = self.exCouponCalendar_.advanceDate(paymentDate,
                                                                  -self.exCouponPeriod_,
                                                                  self.exCouponAdjustment_,
                                                                  self.exCouponEndOfMonth_)

            if (i - 1) < len(self.couponRates_):
                rate = self.couponRates_[i - 1]
            else:
                rate = self.couponRates_[-1]
            if (i - 1) < len(self.notionals_):
                nominal = self.notionals_[i - 1]
            else:
                nominal = self.notionals_[-1]
            leg.append(FixedRateCoupon(paymentDate,
                                       nominal,
                                       rate,
                                       start,
                                       end,
                                       start,
                                       end,
                                       exCouponDate))

        if self.schedule_.size() > 2:
            # last period might be short or long
            N = self.schedule_.size()
            start = end
            end = self.schedule_._dates[N - 1]
            paymentDate = self.calendar_.adjustDate(end, self.paymentAdjustment_)
            if self.exCouponPeriod_:
                exCouponDate = self.exCouponCalendar_.advanceDate(paymentDate,
                                                                  -self.exCouponPeriod_,
                                                                  self.exCouponAdjustment_,
                                                                  self.exCouponEndOfMonth_)
            if (N - 2) < len(self.couponRates_):
                rate = self.couponRates_[N - 2]
            else:
                rate = self.couponRates_[-1]
            if (N - 2) < len(self.notionals_):
                nominal = self.notionals_[N - 2]
            else:
                nominal = self.notionals_[-1]

            if self.schedule_.isRegular(N - 1):
                leg.append(FixedRateCoupon(paymentDate,
                                           nominal,
                                           rate,
                                           start,
                                           end,
                                           start,
                                           end,
                                           exCouponDate))
            else:
                ref = self.schedule_.calendar().advance(start,
                                                        self.schedule_.tenor(),
                                                        self.schedule_._convention,
                                                        self.schedule_.endOfMonth())
                leg.append(FixedRateCoupon(paymentDate,
                                           nominal,
                                           rate,
                                           start,
                                           end,
                                           start,
                                           ref,
                                           exCouponDate))

        return leg



# -*- coding: utf-8 -*-
u"""
Created on 2015-9-23

@author: cheng.li
"""


import functools

from PyFin.Analysis.TechnicalAnalysis import SecurityMovingAverage
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingDecay
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingMax
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingArgMax
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingMin
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingArgMin
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingRank
from PyFin.Analysis.TechnicalAnalysis import SecurityMaximumValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityMinimumValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingQuantile
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingCount
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingCountUnique
from PyFin.Analysis.TechnicalAnalysis import SecurityTimeMovingCount
from PyFin.Analysis.TechnicalAnalysis import SecurityTimeMovingCountUnique
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingAllTrue
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingAnyTrue
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingSum
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingVariance
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingStandardDeviation
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingCountedPositive
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingPositiveAverage
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingRSI
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingCorrelation
from PyFin.Analysis.TechnicalAnalysis import SecuritySignValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityAverageValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityXAverageValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityMACDValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecuritySqrtValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityDiffValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecuritySimpleReturnValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityLogReturnValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityExpValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityLogValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityPowValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityAbsValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityAcosValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityAcoshValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityAsinValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityAsinhValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityNormInvValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityCeilValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityFloorValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityRoundValueHolder
from PyFin.Analysis.SecurityValueHolders import SecurityShiftedValueHolder
from PyFin.Analysis.SecurityValueHolders import SecurityDeltaValueHolder
from PyFin.Analysis.SecurityValueHolders import SecurityCurrentValueHolder
from PyFin.Analysis.SecurityValueHolders import SecurityLatestValueHolder
from PyFin.Analysis.SecurityValueHolders import SecurityIIFValueHolder

from PyFin.Analysis.CrossSectionValueHolders import CSRankedSecurityValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSTopNSecurityValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSBottomNSecurityValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSTopNPercentileSecurityValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSBottomNPercentileSecurityValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSAverageSecurityValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSAverageAdjustedSecurityValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSZScoreSecurityValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSFillNASecurityValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSPercentileSecurityValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSResidueSecurityValueHolder


def CSRank(x, groups=None):
    return CSRankedSecurityValueHolder(x, groups)


def CSTopN(x, n, groups=None):
    return CSTopNSecurityValueHolder(x, n, groups)


def CSTopNQuantile(x, n, groups=None):
    return CSTopNPercentileSecurityValueHolder(x, n, groups)


def CSBottomN(x, n, groups=None):
    return CSBottomNSecurityValueHolder(x, n, groups)


def CSBottomNQuantile(x, n, groups=None):
    return CSBottomNPercentileSecurityValueHolder(x, n, groups)


def CSMean(x, groups=None):
    return CSAverageSecurityValueHolder(x, groups)


def CSMeanAdjusted(x, groups=None):
    return CSAverageAdjustedSecurityValueHolder(x, groups)


def CSQuantiles(x, groups=None):
    return CSPercentileSecurityValueHolder(x, groups)


def CSZScore(x, groups=None):
    return CSZScoreSecurityValueHolder(x, groups)


def CSFillNA(x, groups=None):
    return CSFillNASecurityValueHolder(x, groups)


def CSRes(left, right):
    return CSResidueSecurityValueHolder(left, right)


def SIGN(x='x'):
    return SecuritySignValueHolder(x)


def AVG(x='x'):
    return SecurityAverageValueHolder(x)


def EMA(window, x='x'):
    return SecurityXAverageValueHolder(window, x)


def MACD(short, long, x='x'):
    return SecurityMACDValueHolder(short, long, x)


def RSI(window, x='x'):
    return SecurityMovingRSI(window, x)


def MCORR(window, x='x', y='y'):
    return SecurityMovingCorrelation(window, x, y)


def MA(window, x='x'):
    return SecurityMovingAverage(window, x)


def MADecay(window, x='x'):
    return SecurityMovingDecay(window, x)


def MMAX(window, x='x'):
    return SecurityMovingMax(window, x)


def MARGMAX(window, x='x'):
    return SecurityMovingArgMax(window, x)


def MMIN(window, x='x'):
    return SecurityMovingMin(window, x)


def MARGMIN(window, x='x'):
    return SecurityMovingArgMin(window, x)


def MRANK(window, x='x'):
    return SecurityMovingRank(window, x)


def MAXIMUM(x='x', y='y'):
    return SecurityMaximumValueHolder(x, y)


def MINIMUM(x='x', y='y'):
    return SecurityMinimumValueHolder(x, y)


def MQUANTILE(window, x='x'):
    return SecurityMovingQuantile(window, x)


def MCOUNT(window, x="x", closed="right"):
    if isinstance(window, int):
        return SecurityMovingCount(window, x)
    elif isinstance(window, str):
        return SecurityTimeMovingCount(window, x, closed)
    else:
        raise ValueError("wrong window format <{0}>".format(window))


def MUCOUNT(window, x="x", closed="right"):
    if isinstance(window, int):
        return SecurityMovingCountUnique(window, x)
    elif isinstance(window, str):
        return SecurityTimeMovingCountUnique(window, x, closed)
    else:
        raise ValueError("wrong window format <{0}>".format(window))


def MALLTRUE(window, x='x'):
    return SecurityMovingAllTrue(window, x)


def MANYTRUE(window, x='x'):
    return SecurityMovingAnyTrue(window, x)


def MSUM(window, x='x'):
    return SecurityMovingSum(window, x)


def MVARIANCE(window, x='x'):
    return SecurityMovingVariance(window, x)


def MSTD(window, x='x'):
    return SecurityMovingStandardDeviation(window, x)


def MNPOSITIVE(window, x='x'):
    return SecurityMovingCountedPositive(window, x)


def MAPOSITIVE(window, x='x'):
    return SecurityMovingPositiveAverage(window, x)


def CURRENT(x='x'):
    return SecurityCurrentValueHolder(x)


def LAST(x='x'):
    return SecurityLatestValueHolder(x)


def SQRT(x='x'):
    return SecuritySqrtValueHolder(x)


def DIFF(x='x'):
    return SecurityDiffValueHolder(x)


def RETURNSimple(x='x'):
    return SecuritySimpleReturnValueHolder(x)


def RETURNLog(x='x'):
    return SecurityLogReturnValueHolder(x)


def EXP(x):
    return SecurityExpValueHolder(x)


def LOG(x):
    return SecurityLogValueHolder(x)


def POW(x, n):
    return SecurityPowValueHolder(x, n)


def ABS(x):
    return SecurityAbsValueHolder(x)


def ACOS(x):
    return SecurityAcosValueHolder(x)


def ACOSH(x):
    return SecurityAcoshValueHolder(x)


def ASIN(x):
    return SecurityAsinValueHolder(x)


def ASINH(x):
    return SecurityAsinhValueHolder(x)


def NORMINV(x):
    return SecurityNormInvValueHolder(x)


def CEIL(x):
    return SecurityCeilValueHolder(x)


def FLOOR(x):
    return SecurityFloorValueHolder(x)


def ROUND(x):
    return SecurityRoundValueHolder(x)


def SHIFT(x, n):
    return SecurityShiftedValueHolder(n, x)


def DELTA(x, n):
    return SecurityDeltaValueHolder(n, x)


def IIF(flag, left, right):
    return SecurityIIFValueHolder(flag, left, right)


def INDUSTRY(name, level=1):
    name = name.lower()
    if name == 'sw':
        if level in (1, 2, 3):
            return LAST(name + str(level))
        else:
            raise ValueError('{0} is not recognized level for {1}'.format(level, name))
    else:
        raise ValueError('currently only `sw` industry is supported. {1} is not a recognized industry type')


HIGH = functools.partial(LAST, 'highestPrice')
LOW = functools.partial(LAST, 'lowestPrice')
OPEN = functools.partial(LAST, 'openPrice')
CLOSE = functools.partial(LAST, 'closePrice')






# -*- coding: utf-8 -*-
u"""
Created on 2015-7-13

@author: cheng.li
"""

from PyFin.api.DateUtilities import datesList
from PyFin.api.DateUtilities import bizDatesList
from PyFin.api.DateUtilities import holDatesList
from PyFin.api.DateUtilities import advanceDate
from PyFin.api.DateUtilities import BizDayConventions
from PyFin.api.DateUtilities import DateGeneration
from PyFin.api.DateUtilities import isBizDay
from PyFin.api.DateUtilities import adjustDateByCalendar
from PyFin.api.DateUtilities import advanceDateByCalendar
from PyFin.api.DateUtilities import nthWeekDay
from PyFin.api.DateUtilities import makeSchedule

from PyFin.Analysis import transform
from PyFin.Analysis.SeriesValues import SeriesValues
from PyFin.api.Analysis import SIGN
from PyFin.api.Analysis import AVG
from PyFin.api.Analysis import EMA
from PyFin.api.Analysis import MACD
from PyFin.api.Analysis import RSI
from PyFin.api.Analysis import MCORR
from PyFin.api.Analysis import MA
from PyFin.api.Analysis import MADecay
from PyFin.api.Analysis import MMAX
from PyFin.api.Analysis import MARGMAX
from PyFin.api.Analysis import MMIN
from PyFin.api.Analysis import MARGMIN
from PyFin.api.Analysis import MRANK
from PyFin.api.Analysis import MAXIMUM
from PyFin.api.Analysis import MINIMUM
from PyFin.api.Analysis import MQUANTILE
from PyFin.api.Analysis import MCOUNT
from PyFin.api.Analysis import MUCOUNT
from PyFin.api.Analysis import MALLTRUE
from PyFin.api.Analysis import MANYTRUE
from PyFin.api.Analysis import MSUM
from PyFin.api.Analysis import MVARIANCE
from PyFin.api.Analysis import MSTD
from PyFin.api.Analysis import MNPOSITIVE
from PyFin.api.Analysis import MAPOSITIVE
from PyFin.api.Analysis import CURRENT
from PyFin.api.Analysis import LAST
from PyFin.api.Analysis import HIGH
from PyFin.api.Analysis import LOW
from PyFin.api.Analysis import OPEN
from PyFin.api.Analysis import CLOSE
from PyFin.api.Analysis import SQRT
from PyFin.api.Analysis import DIFF
from PyFin.api.Analysis import RETURNSimple
from PyFin.api.Analysis import RETURNLog
from PyFin.api.Analysis import EXP
from PyFin.api.Analysis import LOG
from PyFin.api.Analysis import POW
from PyFin.api.Analysis import ABS
from PyFin.api.Analysis import ACOS
from PyFin.api.Analysis import ACOSH
from PyFin.api.Analysis import ASIN
from PyFin.api.Analysis import ASINH
from PyFin.api.Analysis import NORMINV
from PyFin.api.Analysis import CEIL
from PyFin.api.Analysis import FLOOR
from PyFin.api.Analysis import ROUND
from PyFin.api.Analysis import SHIFT
from PyFin.api.Analysis import IIF
from PyFin.api.Analysis import INDUSTRY

from PyFin.api.Analysis import CSRank
from PyFin.api.Analysis import CSTopN
from PyFin.api.Analysis import CSBottomN
from PyFin.api.Analysis import CSTopNQuantile
from PyFin.api.Analysis import CSBottomNQuantile
from PyFin.api.Analysis import CSMean
from PyFin.api.Analysis import CSMeanAdjusted
from PyFin.api.Analysis import CSQuantiles
from PyFin.api.Analysis import CSZScore
from PyFin.api.Analysis import CSFillNA
from PyFin.api.Analysis import CSRes

from PyFin.Utilities.Asserts import require


__all__ = ["datesList",
           "bizDatesList",
           "holDatesList",
           "isBizDay",
           "advanceDate",
           "BizDayConventions",
           "DateGeneration",
           "adjustDateByCalendar",
           "advanceDateByCalendar",
           "nthWeekDay",
           "makeSchedule",
           "transform",
           "SIGN",
           "SeriesValues",
           "AVG",
           "EMA",
           "MACD",
           "RSI",
           "MCORR",
           "MA",
           "MADecay",
           "MMAX",
           "MARGMAX",
           "MMIN",
           "MARGMIN",
           "MRANK",
           "MAXIMUM",
           "MINIMUM",
           "MQUANTILE",
           "MCOUNT",
           "MUCOUNT",
           "MALLTRUE",
           "MANYTRUE",
           "MSUM",
           "MVARIANCE",
           "MSTD",
           "MNPOSITIVE",
           "MAPOSITIVE",
           "CURRENT",
           "LAST",
           "HIGH",
           "LOW",
           "OPEN",
           "CLOSE",
           "SQRT",
           "DIFF",
           "RETURNSimple",
           "RETURNLog",
           "EXP",
           "LOG",
           "POW",
           "ABS",
           "ACOS",
           "ACOSH",
           "ASIN",
           "ASINH",
           "NORMINV",
           "CEIL",
           "FLOOR",
           "ROUND",
           "SHIFT",
           "IIF",
           "INDUSTRY",
           "CSRank",
           "CSTopN",
           "CSBottomN",
           "CSTopNQuantile",
           "CSBottomNQuantile",
           "CSMean",
           "CSMeanAdjusted",
           "CSQuantiles",
           "CSZScore",
           "CSFillNA",
           "CSRes",
           "require"]



# -*- coding: utf-8 -*-
u"""
Created on 2015-8-8

@author: cheng.li
"""

from PyFin.Analysis.DataProviders import DataProvider
from PyFin.Analysis.SecurityValueHolders import SecurityShiftedValueHolder
from PyFin.Analysis.SecurityValueHolders import SecurityDeltaValueHolder
from PyFin.Analysis.SecurityValueHolders import SecurityIIFValueHolder
from PyFin.Analysis.SecurityValueHolders import SecurityConstArrayValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSRankedSecurityValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSTopNSecurityValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSBottomNSecurityValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSAverageSecurityValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSAverageAdjustedSecurityValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSZScoreSecurityValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSFillNASecurityValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSPercentileSecurityValueHolder
from PyFin.Analysis.CrossSectionValueHolders import CSResidueSecurityValueHolder
from PyFin.Analysis.SecurityValueHolders import SecurityCurrentValueHolder
from PyFin.Analysis.SecurityValueHolders import SecurityLatestValueHolder
from PyFin.Analysis import TechnicalAnalysis
from PyFin.Analysis.transformer import transform

__all__ = ['DataProvider',
           'SecurityShiftedValueHolder',
           'SecurityDeltaValueHolder',
           'SecurityIIFValueHolder',
           'SecurityConstArrayValueHolder',
           'CSRankedSecurityValueHolder',
           'CSTopNSecurityValueHolder',
           'CSBottomNSecurityValueHolder',
           'CSAverageSecurityValueHolder',
           'CSAverageAdjustedSecurityValueHolder',
           'CSZScoreSecurityValueHolder',
           'CSFillNASecurityValueHolder',
           'CSPercentileSecurityValueHolder',
           'CSResidueSecurityValueHolder',
           'SecurityCurrentValueHolder',
           'SecurityLatestValueHolder',
           'TechnicalAnalysis',
           'transform']



# -*- coding: utf-8 -*-
u"""
Created on 2017-6-11

@author: cheng.li
"""

import unittest
from PyFin.Enums.Frequencies import Frequency
from PyFin.Enums.Compoundings import Compounding
from PyFin.CashFlows.InterestRate import InterestRate
from PyFin.DateUtilities.Date import Date
from PyFin.DateUtilities.DayCounter import DayCounter


class InterestRateData(object):
    def __init__(self, r, comp, freq, t, comp2, freq2, expected, precision):
        self.r = r
        self.comp = comp
        self.freq = freq
        self.t = t
        self.comp2 = comp2
        self.freq2 = freq2
        self.expected = expected
        self.precision = precision


class TestInterestRate(unittest.TestCase):
    def setUp(self):
        self.cases = [
            # data from "Option Pricing Formulas", Haug, pag .181 - 182
            # Rate, Compounding, Frequency, Time, Compounding2, Frequency2, Rate2, precision
            InterestRateData(0.0800, Compounding.Compounded, Frequency.Quarterly, 1.00, Compounding.Continuous,
                             Frequency.Annual, 0.0792, 4),
            InterestRateData(0.1200, Compounding.Continuous, Frequency.Annual, 1.00, Compounding.Compounded,
                             Frequency.Annual, 0.1275, 4),
            InterestRateData(0.0800, Compounding.Compounded, Frequency.Quarterly, 1.00, Compounding.Compounded,
                             Frequency.Annual, 0.0824, 4),
            InterestRateData(0.0700, Compounding.Compounded, Frequency.Quarterly, 1.00, Compounding.Compounded,
                             Frequency.Semiannual, 0.0706, 4),
            # undocumented, but reasonable:)
            InterestRateData(0.0100, Compounding.Compounded, Frequency.Annual, 1.00, Compounding.Simple,
                             Frequency.Annual, 0.0100, 4),
            InterestRateData(0.0200, Compounding.Simple, Frequency.Annual, 1.00, Compounding.Compounded,
                             Frequency.Annual, 0.0200, 4),
            InterestRateData(0.0300, Compounding.Compounded, Frequency.Semiannual, 0.50, Compounding.Simple,
                             Frequency.Annual, 0.0300, 4),
            InterestRateData(0.0400, Compounding.Simple, Frequency.Annual, 0.50, Compounding.Compounded,
                             Frequency.Semiannual, 0.0400, 4),
            InterestRateData(0.0500, Compounding.Compounded, Frequency.EveryFourthMonth, 1.0 / 3, Compounding.Simple,
                             Frequency.Annual, 0.0500, 4),
            InterestRateData(0.0600, Compounding.Simple, Frequency.Annual, 1.0 / 3, Compounding.Compounded,
                             Frequency.EveryFourthMonth, 0.0600, 4),
            InterestRateData(0.0500, Compounding.Compounded, Frequency.Quarterly, 0.25, Compounding.Simple,
                             Frequency.Annual, 0.0500, 4),
            InterestRateData(0.0600, Compounding.Simple, Frequency.Annual, 0.25, Compounding.Compounded,
                             Frequency.Quarterly, 0.0600, 4),
            InterestRateData(0.0700, Compounding.Compounded, Frequency.Bimonthly, 1.0 / 6, Compounding.Simple,
                             Frequency.Annual, 0.0700, 4),
            InterestRateData(0.0800, Compounding.Simple, Frequency.Annual, 1.0 / 6, Compounding.Compounded,
                             Frequency.Bimonthly, 0.0800, 4),
            InterestRateData(0.0900, Compounding.Compounded, Frequency.Monthly, 1.0 / 12, Compounding.Simple,
                             Frequency.Annual, 0.0900, 4),
            InterestRateData(0.1000, Compounding.Simple, Frequency.Annual, 1.0 / 12, Compounding.Compounded,
                             Frequency.Monthly, 0.1000, 4),

            InterestRateData(0.0300, Compounding.SimpleThenCompounded, Frequency.Semiannual, 0.25, Compounding.Simple,
                             Frequency.Annual, 0.0300, 4),
            InterestRateData(0.0300, Compounding.SimpleThenCompounded, Frequency.Semiannual, 0.25, Compounding.Simple,
                             Frequency.Semiannual, 0.0300, 4),
            InterestRateData(0.0300, Compounding.SimpleThenCompounded, Frequency.Semiannual, 0.25, Compounding.Simple,
                             Frequency.Quarterly, 0.0300, 4),
            InterestRateData(0.0300, Compounding.SimpleThenCompounded, Frequency.Semiannual, 0.50, Compounding.Simple,
                             Frequency.Annual, 0.0300, 4),
            InterestRateData(0.0300, Compounding.SimpleThenCompounded, Frequency.Semiannual, 0.50, Compounding.Simple,
                             Frequency.Semiannual, 0.0300, 4),
            InterestRateData(0.0300, Compounding.SimpleThenCompounded, Frequency.Semiannual, 0.75,
                             Compounding.Compounded, Frequency.Semiannual, 0.0300, 4),

            InterestRateData(0.0400, Compounding.Simple, Frequency.Semiannual, 0.25, Compounding.SimpleThenCompounded,
                             Frequency.Quarterly, 0.0400, 4),
            InterestRateData(0.0400, Compounding.Simple, Frequency.Semiannual, 0.25, Compounding.SimpleThenCompounded,
                             Frequency.Semiannual, 0.0400, 4),
            InterestRateData(0.0400, Compounding.Simple, Frequency.Semiannual, 0.25, Compounding.SimpleThenCompounded,
                             Frequency.Annual, 0.0400, 4),

            InterestRateData(0.0400, Compounding.Compounded, Frequency.Quarterly, 0.50,
                             Compounding.SimpleThenCompounded, Frequency.Quarterly, 0.0400, 4),
            InterestRateData(0.0400, Compounding.Simple, Frequency.Semiannual, 0.50, Compounding.SimpleThenCompounded,
                             Frequency.Semiannual, 0.0400, 4),
            InterestRateData(0.0400, Compounding.Simple, Frequency.Semiannual, 0.50, Compounding.SimpleThenCompounded,
                             Frequency.Annual, 0.0400, 4),

            InterestRateData(0.0400, Compounding.Compounded, Frequency.Quarterly, 0.75,
                             Compounding.SimpleThenCompounded, Frequency.Quarterly, 0.0400, 4),
            InterestRateData(0.0400, Compounding.Compounded, Frequency.Semiannual, 0.75,
                             Compounding.SimpleThenCompounded, Frequency.Semiannual, 0.0400, 4),
            InterestRateData(0.0400, Compounding.Simple, Frequency.Semiannual, 0.75, Compounding.SimpleThenCompounded,
                             Frequency.Annual, 0.0400, 4)
        ]

    def testConversion(self):

        d1 = Date.todaysDate()

        for case in self.cases:
            ir = InterestRate(case.r, DayCounter('Actual/360'), case.comp, case.freq)
            d2 = d1 + int(360 * case.t + 0.5)

            compoundf = ir.compoundFactor(d1, d2)
            disc = ir.discountFactor(d1, d2)

            self.assertAlmostEqual(compoundf, 1.0 / disc)

            ir2 = ir.equivalentRate(ir.dayCounter_,
                                    ir.comp_,
                                    ir.freq_,
                                    d1, d2)

            self.assertAlmostEqual(ir2.r_, ir.r_)
            self.assertEqual(ir.freq_, ir2.freq_)
            self.assertEqual(ir.comp_, ir2.comp_)

            ir3 = ir.equivalentRate(ir.dayCounter_,
                                    case.comp2, case.freq2,
                                    d1, d2)
            expectedIR = InterestRate(case.expected, ir.dayCounter_,
                                      case.comp2, case.freq2)
            self.assertAlmostEqual(ir3.r_, expectedIR.r_, case.precision)

            r3 = ir.equivalentRate(ir.dayCounter_,
                                   case.comp2, case.freq2,
                                   d1, d2)
            self.assertAlmostEqual(r3.r_, case.expected, case.precision)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
u"""
Created on 2017-1-6

@author: cheng.li
"""

import unittest
import numpy as np
from PyFin.PricingEngines import sabrVolatility
from PyFin.PricingEngines import sabrVolatilities
from PyFin.PricingEngines import sabrCalibration


class TestSabrFormula(unittest.TestCase):

    def setUp(self):
        self.strikes = np.array([0.] * 31)
        self.volatilities = np.array([0.] * 31)

        self.strikes[0] = 0.03
        self.strikes[1] = 0.032
        self.strikes[2] = 0.034
        self.strikes[3] = 0.036
        self.strikes[4] = 0.038
        self.strikes[5] = 0.04
        self.strikes[6] = 0.042
        self.strikes[7] = 0.044
        self.strikes[8] = 0.046
        self.strikes[9] = 0.048
        self.strikes[10] = 0.05
        self.strikes[11] = 0.052
        self.strikes[12] = 0.054
        self.strikes[13] = 0.056
        self.strikes[14] = 0.058
        self.strikes[15] = 0.06
        self.strikes[16] = 0.062
        self.strikes[17] = 0.064
        self.strikes[18] = 0.066
        self.strikes[19] = 0.068
        self.strikes[20] = 0.07
        self.strikes[21] = 0.072
        self.strikes[22] = 0.074
        self.strikes[23] = 0.076
        self.strikes[24] = 0.078
        self.strikes[25] = 0.08
        self.strikes[26] = 0.082
        self.strikes[27] = 0.084
        self.strikes[28] = 0.086
        self.strikes[29] = 0.088
        self.strikes[30] = 0.09

        self.volatilities[0] = 1.16725837321531
        self.volatilities[1] = 1.15226075991385
        self.volatilities[2] = 1.13829711098834
        self.volatilities[3] = 1.12524190877505
        self.volatilities[4] = 1.11299079244474
        self.volatilities[5] = 1.10145609357162
        self.volatilities[6] = 1.09056348513411
        self.volatilities[7] = 1.08024942745106
        self.volatilities[8] = 1.07045919457758
        self.volatilities[9] = 1.06114533019077
        self.volatilities[10] = 1.05226642581503
        self.volatilities[11] = 1.04378614411707
        self.volatilities[12] = 1.03567243073732
        self.volatilities[13] = 1.0278968727451
        self.volatilities[14] = 1.02043417226345
        self.volatilities[15] = 1.01326171139321
        self.volatilities[16] = 1.00635919013311
        self.volatilities[17] = 0.999708323124949
        self.volatilities[18] = 0.993292584155381
        self.volatilities[19] = 0.987096989695393
        self.volatilities[20] = 0.98110791455717
        self.volatilities[21] = 0.975312934134512
        self.volatilities[22] = 0.969700688771689
        self.volatilities[23] = 0.964260766651027
        self.volatilities[24] = 0.958983602256592
        self.volatilities[25] = 0.953860388001395
        self.volatilities[26] = 0.948882997029509
        self.volatilities[27] = 0.944043915545469
        self.volatilities[28] = 0.939336183299237
        self.volatilities[29] = 0.934753341079515
        self.volatilities[30] = 0.930289384251337

        self.expiry = 1.0
        self.forward = 0.039
        self.initialAlpha = 0.3
        self.initialBeta = 0.6
        self.initialNu = 0.02
        self.initialRho = 0.01

    def testSabrVolatility(self):
        for i, strike in enumerate(self.strikes):
            calculatedVol = sabrVolatility(strike,
                                           self.forward,
                                           self.expiry,
                                           self.initialAlpha,
                                           self.initialBeta,
                                           self.initialNu,
                                           self.initialRho)
            expectedVol = self.volatilities[i]
            self.assertAlmostEqual(calculatedVol, expectedVol)

    def testSabrVoaltilities(self):
        calculatedVols = sabrVolatilities(self.strikes,
                                          self.forward,
                                          self.expiry,
                                          self.initialAlpha,
                                          self.initialBeta,
                                          self.initialNu,
                                          self.initialRho)
        expectedVols = self.volatilities
        np.testing.assert_array_almost_equal(calculatedVols, expectedVols)

    def testSabrCalibrationWithTrf(self):
        x = sabrCalibration(self.strikes, self.volatilities, self.forward, self.expiry, 0.01, 0.01, 0.01, 0.01)
        calibratedParameters = x[0]

        calculatedVols = sabrVolatilities(self.strikes,
                                          self.forward,
                                          self.expiry,
                                          calibratedParameters[0],
                                          calibratedParameters[1],
                                          calibratedParameters[2],
                                          calibratedParameters[3])
        expectedVols = self.volatilities
        np.testing.assert_array_almost_equal(calculatedVols, expectedVols, 5)

    def testSabrCalibrationWithLM(self):
        x = sabrCalibration(self.strikes,
                            self.volatilities,
                            self.forward,
                            self.expiry,
                            0.01,
                            0.01,
                            0.01,
                            0.01,
                            method='lm')
        calibratedParameters = x[0]

        calculatedVols = sabrVolatilities(self.strikes,
                                          self.forward,
                                          self.expiry,
                                          calibratedParameters[0],
                                          calibratedParameters[1],
                                          calibratedParameters[2],
                                          calibratedParameters[3])
        expectedVols = self.volatilities
        np.testing.assert_array_almost_equal(calculatedVols, expectedVols, 4)

    def testSabrCalibrationWithFixedBeta(self):
        x = sabrCalibration(self.strikes,
                            self.volatilities,
                            self.forward,
                            self.expiry,
                            0.01,
                            self.initialBeta,
                            0.01,
                            0.01,
                            isFixedBeta=True)
        calibratedParameters = x[0]

        self.assertAlmostEqual(calibratedParameters[1], self.initialBeta)

        calculatedVols = sabrVolatilities(self.strikes,
                                          self.forward,
                                          self.expiry,
                                          calibratedParameters[0],
                                          calibratedParameters[1],
                                          calibratedParameters[2],
                                          calibratedParameters[3])

        expectedVols = self.volatilities
        np.testing.assert_array_almost_equal(calculatedVols, expectedVols, 4)
